# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
Root package of :py:mod:`imprint`.

This package contains the following sub-packages:

  - :py:mod:`~imprint.core`:
      The code that generates the documents. This includes all XML
      parsers and implementations of the :ref:`tag-api-builtins`.
  - :py:mod:`~imprint.handlers`:
      A package implementing common :ref:`plugins-handlers` for the
      built-in tags that accept them.
  - :py:mod:`~imprint.tests`:
      While woefully incomplete, this package contains the beginnings of 
      unit tests for imprint.

The public driver script for the :py:mod:`imprint` package is
:program:`imprint`, implemented in the module
:py:mod:`imprint.driver`.
"""


__all__ = ['__version__']



from .version import __version__
"""
The current version of the software. Individual components, especially in the
configuration, may have different version numbers.
"""



# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
:py:mod:`imprint.handlers.table` is the root package for built-in
:ref:`plugins-handlers` for inserting tables into a document.

All the handlers in this module are compatible with the
:ref:`plugin interface <plugins-tables-signature>` used by the
:ref:`xml-spec-tags-table` tag. This package exposes all the handlers
defined in its submodules.

Some of the handlers in this package depend on the ``[all]`` or
``[table]`` extra being :ref:`installed <installation-install>`. An
alternative way to enable these features is to manually install the
`pandas`_ library in your environment.
"""


from itertools import repeat

from docx.enum.table import WD_TABLE_ALIGNMENT
from docx.shared import Emu

from haggis.files.docx import merge_row, set_row_height, str2length


__all__ = ['fixed_width_table',]


#: Mapping of valid alignment values for :py:func:fixed_width_table` to
#: The corresponding :py:class:`~docx.enum.table.WD_TABLE_ALIGNMENT`
#: values. The keys of this mapping are valid values for the
#: ``alignment`` key used in the :ref:`plugins-data-configuration` of
#: the :py:func:`~imprint.core.handler.table.tables.CSVFile` and
#: :py:func:`~imprint.core.handler.table.tables.DataFrame` handlers.
_table_alignment = {
    'left': WD_TABLE_ALIGNMENT.LEFT,
    'right': WD_TABLE_ALIGNMENT.RIGHT,
    'center': WD_TABLE_ALIGNMENT.CENTER,
    '<' : WD_TABLE_ALIGNMENT.LEFT,
    '>': WD_TABLE_ALIGNMENT.RIGHT,
    '^': WD_TABLE_ALIGNMENT.CENTER,
}


def fixed_width_table(doc, style, rows=1, cols=None, alignment='^',
                      width=None, col_widths=None, row_height=None):
    """
    Construct a table of fixed width, with optionally specified column
    widths.

    Parameters
    ----------
    doc : docx.document.Document
        The document to insert the table into.
    style : str
        The name of the style to assign to the document.
    rows : int
        The number of rows to add to the table.
    cols : int or None
        The number of colums to add to the table. If :py:obj:`None`, the
        length of ``col_widths`` is used. Either this argument or
        ``col_widths`` must be non-falsy.
    alignment : { '^', '<', '>' }
        The direction in which the table shoule be aligned. Defaults to
        center (``'^'``).
    width : length-spec
        Either a number in inches, or a string representing a number
        with optional units. A full list of accepted units is provided
        in :py:func:`haggis.files.docx.str2length`. Ignored if
        ``col_widths`` is supplied in a way that does not leave
        unspecified columns. Defaults to 7 inches.
    col_widths : sequence[length-spec]
        A sequence of widths, each in the same format as ``width``, for
        the individual columns of the table. If this is longer than
        ``cols``, only the required number of elements is used. If
        shorter, the remaining columns are divided evenly to fill
        ``width``.
    row_height : None or length-spec or sequence[length-spec]
        The height to assign to each row. :py:obj:`None` is the default,
        meaning to let the rows adjust themselves. If a single number or
        numerical string with optional units, it will be applied to all
        the newly created rows. If a sequence, only the minimum of
        ``rows`` and the length of the sequence will be affected. A
        sequence may contain any of the values allowed as scalars
        (including :py:obj:`None`).

    Return
    ------
    table : docx.table.Table
        The newly created table object.
    """
    if width is None:
        width = 7.0
    width = str2length(width)

    if col_widths is None:
        col_widths = []
    else:
        col_widths = [str2length(w) for w in col_widths]

    if not cols:
        if not col_widths:
            raise('Either the number of columns or their widths must be set')
        cols = len(col_widths)

    table = doc.add_table(rows=rows, cols=0, style=style)
    try:
        table.alignment = _table_alignment[alignment]
    except KeyError:
        raise ValueError(f'Illegal value of alignment {alignment!r}')
    table.autofit = not col_widths

    if cols > len(col_widths):
        extra_cols = cols - len(col_widths)
        total = Emu(sum(w.emu for w in col_widths))
        if total > width:
            raise ValueError(f'Existing columns occupy {total.inches}" '
                             f'of {total.width}".')
        delta = width.emu - total.emu
        col_widths += [Emu(delta / extra_cols)] * (extra_cols - 1) + \
                      [Emu(delta / extra_cols + delta % extra_cols)]
    else:
        col_widths = col_widths[:cols]

    for w in col_widths:
        table.add_column(w)

    try:
        iter(row_height)
    except:
        row_height = repeat(row_height)
    else:
        if isinstance(row_height, str):
            row_height = repeat(row_height)

    for row, height in zip(table.rows, row_height):
        if height is None:
            continue
        height = str2length(height)
        set_row_height(row, height)

    return table


from .tables import *
from .tables import __all__ as _all
__all__.extend(_all)
del _all



# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
:py:mod:`imprint.handlers.string.strings` contains the basic built-in
:ref:`plugins-handlers` for inserting strings into a document. All the
handlers in this module are compatible with the
:ref:`plugin interface <plugins-strings-signature>` used by the
:ref:`xml-spec-tags-string` tag.
"""

import logging

from ..utilities import get_file


__all__ = ['TextFile',]


#: Private logger for this module
_logger = logging.getLogger(__name__)


def TextFile(config, kwds):
    """
    Generate a string directly from the contents of a text file.

    Text files are inserted literally, with no styling information
    beyond that of the :ref:`xml-spec-tags-string` tag that triggered
    the plugin. Newlines are not preserved.

    The following :ref:`plugins-data-configuration` keys are used:

        file
            The (mandatory) file name.
        formatted
            Whether or not ``file`` is a format string that has keyword
            replacements in it. Defaults to truthy. Set to falsy if
            the name contains random opening braces.
    """
    input_path = get_file(config, kwds)
    _logger.trace('Loading string from %r', input_path)
    with open(input_path, 'rt') as file:
        return file.read()



# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
The :py:mod:`imprint.core.state` module supplies the state objects that
enable communication within the :ref:`introduction-layers-engine`
between the engine itself and the tags. The state is therefore crucial
to the :ref:`tag-api` without being completely a part of it.
"""

from collections import deque
from contextlib import contextmanager
from enum import Enum, unique
from itertools import repeat
from io import StringIO
import logging
from operator import index
from os.path import abspath, splitext

from haggis.files.docx import is_paragraph_empty, delete_paragraph, list_number

from . import defaults
from .utilities import aggressive_strip


__all__ = ['ReferenceMap', 'ReferenceState', 'ListType', 'EngineState']


class ReferenceMap:
    """
    A multi-level mapping that stores references in the values.

    Values are accessed through a three-level key
    ``(role, attribute, key)``: For a given
    :ref:`xml-spec-attributes-role`, the type of key is determined by
    the ``attribute`` that names the target. Most tags only support
    ``attribute='id'``, but :ref:`xml-spec-tags-segment-ref` also
    supports ``attribute='title'``. ``key`` is the actual value of the
    attribute that is used to identify the reference.

    Reference values can be any object whose :py:meth:`~object.__str__`
    method returns the correct replacement text for the reference.
    """
    # TODO: Add this class to haggis (as a generic multi-level mapping).
    def __new__(cls, *args, **kwargs):
        """
        Ensure that the map is unlocked when it is first created.

        This way calling :py:meth:`~object.__init__` is **not** a trick
        for unlocking the map.
        """
        self = super().__new__(cls, *args, **kwargs)
        self.__dict__['_locked'] = False
        self.__dict__['_mapping'] = {}
        return self

    @property
    def _locked(self):
        """
        Checks if this mapping is locked.

        This property can be set to :py:obj:`True` at any time. After
        it is set to :py:obj:`True`, it can never be set back to
        :py:obj:`False`, and none of the setter operations will work.
        """
        return self.__dict__['_locked']

    @_locked.setter
    def _locked(self, value):
        """
        Setter for :py:attr:`_locked`.
        """
        if value:
            self.__dict__['_locked'] = True
        elif self._locked:
            cls = type(self).__name__
            raise ValueError(f'Unlocking a locked {cls} is forbidden!')

    @property
    def _mapping(self):
        """
        Hidden attribute.
        """
        cls = type(self).__name__
        raise AttributeError(f'{cls} object has no attribute "_mapping"')

    def __getitem__(self, key):
        """
        Retreive the value for the specified three-level key.
        """
        role, attr, ref = self._check_key(key)
        role_map = self.__dict__['_mapping']
        try:
            attr_map = role_map[role]
        except KeyError:
            raise KeyError(f'No such role {role!r}')
        try:
            ref_map = attr_map[attr]
        except KeyError:
            raise KeyError(f'No such attribute {attr!r} for {role!r}')
        try:
            return ref_map[ref]
        except KeyError:
            raise KeyError(f'{role!r} has no key {ref!r} '
                           f'under attribute {attr!r}')

    def __setitem__(self, key, value):
        """
        If this mapping is not locked, set the attribute for the
        specified three-level key.

        If any of the levels are new, they are created along the
        way.
        """
        if self._locked:
            cls = type(self).__name__
            raise ValueError(f'Unable to set {key!r} on locked {cls}')
        role, attr, ref = self._check_key(key)
        role_map = self.__dict__['_mapping']
        if role not in role_map:
            role_map[role] = {}
        attr_map = role_map[role]
        if attr not in attr_map:
            attr_map[attr] = {}
        ref_map = attr_map[attr]
        ref_map[ref] = value

    def __contains__(self, key):
        """
        Checks if this mapping has the specified partial key.

        Key may be a single string or a :py:class:`tuple` with a length
        between 1 and 3. Checks will be made for the appropriate depth.
        """
        role_map = self.__dict__['_mapping']
        if isinstance(key, str):
            return key in role_map
        elif not isinstance(key, tuple) or 3 < len(key) <= 0:
            raise TypeError('Containment check only allowed for role (str) '
                            'or partial keys (tuple)')
        role = key[0]
        if role not in role_map:
            return False
        if len(key) == 1:
            return True
        attr_map = role_map[role]
        attr = key[1]
        if attr not in attr_map:
            return False
        if len(key) == 2:
            return True
        ref_map = attr_map[attr]
        ref = key[2]
        return ref in ref_map

    def __str__(self, indent=2):
        """
        Creates a pretty representation of this map, with indented
        heading levels.
        """
        role_map = self.__dict__['_mapping']
        output = StringIO()
        indent1 = ' ' * indent
        indent2 = 2 * indent1
        indent3 = 3 * indent1
        output.write(f'{type(self).__name__}(\n')
        for role, attr_map in role_map.items():
            output.write(f'{indent1}role = {role!r}: {{\n')
            for attr, ref_map in attr_map.items():
                output.write(f'{indent2}attribute = {attr!r}: {{\n')
                for ref, value in ref_map.items():
                    output.write(f'{indent3}key = {ref!r}: "{value}"\n')
                output.write(f'{indent2}}}\n')
            output.write(f'{indent1}}}\n')
        output.write(')\n')
        return output.getvalue()

    def lock(self):
        """
        Lock this mapping to prevent unintentional modification.

        This is a one-time operation. There is no way to unlock.
        After locking, :py:meth:`__setitem__` will raise an error.
        """
        self._locked = True

    @staticmethod
    def _check_key(key):
        """
        Verify that the supplied key is valid to index this map, raise
        a :py:exc:`TypeError` if not.

        This utility just checks that the key is a 3-element
        :py:class:`tuple` of strings.
        """
        if not isinstance(key, tuple) or len(key) != 3:
            raise TypeError('Key of the form (role, attribute, key) required')
        if not all(isinstance(x, str) for x in key):
            raise TypeError('Only strings allowed in key')
        return key


class _StateBase:
    """
    The base container type for the states of the different phases of
    the :ref:`introduction-layers-engine`.

    This class abstracts the common functionality used by all the engine
    components. It allows for a containment check using ``in`` in
    preferece to :py:func:`hasattr`.

    .. py:attribute:: references

       :py:class:`ReferenceMap`

       A multi-level mapping type that allows references to be fetched
       by role and attribute. Access to this map is performed by
       providing a tuple ``(role, attribute, key)``. For example::

           state.references['figure', 'id', 'my_figure']

       The map's values may be of any type, as long as they can be
       converted to the desired content using :py:class:`str`.

       Implemented as a read-only property.
    """
    def __init__(self, references, log):
        self.__dict__.update(
            references=references,
            log=log,
        )

    def __contains__(self, name):
        """
        Checks if the specified name represents an attribute.
        """
        return hasattr(self, name)

    @property
    def references(self):
        """
        Ensure that :py:attr:`references` is read-only.
        """
        return self.__dict__['references']

    def log(self, lvl, msg, *args, **kwargs):
        """
        Provide access to the engine's logging facility.

        Usage is analagous to :py:func:`logging.log`. XML location
        meta-data will be inserted into any log messages.
        """
        # This method is actually just a placeholder for the documentation.
        # It gets overriden by a bound method passed in to the constructor.
        pass

    def get_content(self, default=''):
        """
        Retrieve the text in the current :py:attr:`content` buffer.

        Whitespace is stripped from each line in the text, which is then
        recombined with spaces instead of newlines.

        If the buffer is non-existent, empty or contains only
        whitespace, return `default` instead.
        """
        if 'content' not in self:
            return default
        text = aggressive_strip(self.content.getvalue())
        if not text:
            text = default
        return text


class ReferenceState(_StateBase):
    """
    A simple container type used by the reference parser to communicate
    state to the reference descriptors and accumulate the reference map.

    Most of the state is dedicated to monitoring referenceable tags and
    creating references to them. The engine and built-in tags rely on a
    set of attributes to function properly. A description of acceptable
    use of these attributes is provided here. Any other use may lead to
    unexpected behavior. Custom tags may define and use any attributes
    that are not explicitly documented as they chose.

    This class allows for a containment check using ``in`` in preferece
    to :py:func:`hasattr`.

    .. py:attribute:: registry

       ``Mapping``

       A subtype of :py:class:`dict` that follows the same rules as
       :py:data:`~imprint.core.tags.tag_registry`. Normally a reference
       to that attribute.

       Implemented as a read-only property.

    .. py:attribute:: references

       :py:class:`ReferenceMap`

       A multi-level mapping type that allows references to be fetched
       and set by role and attribute. Access to this map is performed by
       providing a tuple ``(role, attribute, key)``. For example::

           state.references['figure', 'id', 'my_figure']

       The map's values may be of any type, as long as they can be
       converted to the desired content using :py:class:`str`.

       The map is mutable at this stage in the processing. It
       accumulates all the referenceable tags found in the document.
       Setting a value for a key any of whose levels do not exist is
       completely acceptable: the missing levels will be filled in.

       Implemented as a read-only property.

    .. py:attribute:: heading_depth

       :py:class:`int`

       The configured depth after which :py:attr:`heading_counter` stops
       having an effect when a subheading is entered. If omitted
       entirely (:py:obj:`None`), all available heading levels will be
       used.

       Implemented as a writable property.

    .. py:attribute:: heading_counter

       :py:class:`list`\ [\ :py:class:`int`\ ]

       A list containing counters for each heading level encountered.
       The list is popped back one element whenever a higher level
       heading is encountered. ``len(heading_counter)`` is the depth of
       the outline the parser is currently in. E.g., if the parser is
       parsing text under ``Section 3.4.5``, ``heading_counter``
       contains ``[3, 4, 5]``. When ``Section 4`` is encountered next,
       the counter will be reset to ``[4]``. The heading may be
       referenced later by title or by ID.

       A :py:class:`~collections.deque` is not used because it does not
       support slice deletion, which makes jumping back a few heading
       levels much easier.

       Implemented as a read-only property.

    .. py:attribute:: item_counters

       :py:class:`dict`\ [\ :py:class:`str` -> :py:class:`int`\ ]

       A mapping of the :term:referenceable roles to the counters of
       items in the current heading. All the counters are reset to zero
       when a new heading below :py:attr:`heading_depth` is encountered.

       Implemented as a read-only property. The keys of the mapping
       should not be modified, but the values may be.

    .. py:attribute:: content

       :py:class:`io.StringIO`

       A mutable buffer used by the engine to accumulate text from the
       :ref:`configuration-xml` only when necessary.

       This attribute should be manipulated mostly through the
       :py:meth:`start_content` and :py:meth:`end_content` methods. It
       should only be present for tags that care about accumulating
       content for a reference, like :ref:`xml-spec-tags-par`. When
       present, all content, regardless of nested tags, will be
       accumulated.
    """
    def __init__(self, registry, log, heading_depth=None):
        super().__init__(references=ReferenceMap(), log=log)
        self.__dict__.update(
            registry=registry,
            heading_counter=[],
            item_counters={key: 0 for key in registry.referable_tags},
        )
        self.heading_depth = heading_depth

    @property
    def registry(self):
        """
        Ensure that :py:attr:`registry` is read-only.
        """
        return self.__dict__['registry']

    @property
    def heading_counter(self):
        """
        Ensure that :py:attr:`heading_counter` is read-only.
        """
        return self.__dict__['heading_counter']

    @property
    def item_counters(self):
        """
        Ensure that :py:attr:`item_counters` is read-only.
        """
        return self.__dict__['item_counters']

    @property
    def heading_depth(self):
        """
        Ensure that :py:attr:`heading_depth` is set to a legitimate
        value.
        """
        return self.__dict__['heading_depth']

    @heading_depth.setter
    def heading_depth(self, depth):
        if depth is not None:
            depth = index(depth)
            if depth < 0:
                raise ValueError('Negative heading_depth')
        self.__dict__['heading_depth'] = depth

    def start_content(self):
        """
        Create a new :py:attr:`content` buffer.

        If a buffer already exists, a warning is issued (even if it is
        empty), and its contents are discarded.
        """
        if 'content' in self:
            self.log(logging.WARNING, 'Creating duplicate content buffer')
        self.content = StringIO()

    def end_content(self):
        """
        Terminate the current content buffer, if any, and return the
        content after aggressive stripping of whitespace.

        If there is no :py:attr:`content` buffer to begin with, an empty
        string is returned.
        """
        text = self.get_content()
        del self.content
        return text

    def reset_counters(self):
        """
        Set all the values of :py:attr:`item_counters` to zero.
        """
        for key in self.item_counters:
            self.item_counters[key] = 0

    def increment_heading(self, level):
        """
        Increment :py:attr:`heading_counter` at the requested `level`.

        Any missing levels are set to 1 with a warning. Any further
        levels are truncated. :py:attr:`item_counters` is reset if
        :py:attr:`heading_depth` is unset or a greater value than
        `level`.
        """
        if level < 0:
            raise ValueError(f'Negative level {level}')

        self.log(logging.DEBUG, 'Found heading with level %d', level)
        current_level = len(self.heading_counter)

        # Another subheading on the same level
        if level == current_level:
            self.heading_counter[-1] += 1
            self.log(logging.INFO, 'Entered section %s '
                     '(mainained subsection level)', self.format_heading())

        # New subheading: only one level of depth increase allowed at a time
        elif level > current_level:
            diff = level - current_level
            if diff == 1:
                self.heading_counter.append(1)
                self.log(logging.DEBUG, 'Entered section %s '
                         '(subsection level increased to %d)',
                         self.format_heading(), level)
            else:
                self.heading_counter.extend(repeat(1, diff))
                self.log(logging.WARNING, 'Suspicious heading level: '
                         'jumping from depth %d to %d', current_level, level)

        # Jump back to higher level heading
        else:
            del self.heading_counter[level:]
            self.heading_counter[-1] += 1
            self.log(logging.DEBUG, 'Jumped back to parent section %s '
                     '(from depth %d to %d)', self.format_heading(),
                     current_level, level)

        if self.heading_depth is None or level <= self.heading_depth:
            self.reset_counters()
            if self.heading_depth is None:
                msg = 'unlimited depth'
            else:
                msg = f'level={level} <= depth={self.heading_depth}'
            self.log(logging.DEBUG, 'Reset item counters (%s)', msg)
        else:
            self.log(logging.DEBUG, 'Item counters will not be reset '
                     '(level=%d > depth=%d', level, self.heading_depth)

    def format_heading(self, prefix=None, prefix_sep=' ', sep='.',
                       suffix_sep='-', suffix=None):
        """
        Format :py:attr:`heading_counter` for display.

        If `suffix` is set to a Truthy value, only
        :py:attr:`heading_depth` items are shown. Otherwise, the entire
        list is shown.
        """
        depth = self.heading_depth
        if depth is None or not suffix:
            depth = len(self.heading_counter)
        heading = sep.join(map(str, self.heading_counter[:depth]))

        if not prefix:
            prefix = prefix_sep = ''
        if not suffix:
            suffix = suffix_sep = ''
        elif not heading:
            suffix_sep = ''

        return f'{prefix}{prefix_sep}{heading}{suffix_sep}{suffix}'


@unique
class ListType(Enum):
    """
    The type of list numbering to use for :ref:`xml-spec-tags-par` tags
    that require it.
    """

    #: Start a new numbered list.
    NUMBERED = 'numbered'

    #: Start a new bulleted list.
    BULLETED = 'bulleted'

    #: Continue with the numbering/bullets of an existing list.
    CONTINUED = 'continued'

    @classmethod
    def convert(cls, value):
        """
        Normalize the provided value, or raise a :py:exc:`ValueError`.

        Values are assigned based on the case-insensitive prefix that
        they represent. `value` must be a string.
        """
        cvalue = value.casefold()
        for e in cls:
            if e.value.startswith(cvalue):
                return e
        else:
            raise ValueError(f'Illegal value of {cls.__name__} {value!r}')


class EngineState(_StateBase):
    """
    A simple container type used by the main parser to communicate
    document state to the tag descriptors.

    Most of the state is dedicated to monitoring the status of the text
    acquisition from the XML. The engine and built-in tags rely on a set
    of attributes to function. A description of acceptable use of these
    attributes is provided here. Any other use may lead to unexpected
    behavior. Custom tags may define and use any attributes that are not
    explicitly documented as they choose.

    This class allows for a containment check using ``in`` in preferece
    to :py:func:`hasattr`.

    .. py:attribute:: doc

       :py:class:`docx.document.Document`

       The document that is being built. Set once by the engine.

       Implemented as a read-only property.

    .. py:attribute:: keywords

       :py:class:`dict`

       The keywords configured for this document by the
       :ref:`configuration-ipc`. Normally, this dictionary should be
       treated as read-only, but :py:class:`~imprint.core.tags.ExprTag`
       can add new entries.

       As a rule, keywords with lowercase names are system configuration
       options, while keywords that start with upper case letters affect
       document content.

       Implemented as a read-only property.

    .. py:attribute:: references

       :py:class:`ReferenceMap`

       A multi-level mapping type that allows references to be fetched
       by role and attribute. Access to this map is performed by
       providing a tuple ``(role, attribute, key)``. For example::

           state.references['figure', 'id', 'my_figure']

       The map's values may be of any type, as long as they can be
       converted to the desired content using :py:class:`str`.

       The mapping is made immutable as soon as it becomes part of the
       state. The read-only lock is irreversible.

       Implemented as a read-only property.

    .. py:attribute:: paragraph

       :py:class:`docx.text.paragraph.Paragraph`

        A paragraph represents a collection of runs and other objects
        that make up a logical segment in a document. This attribute
        exists only when parsing a :ref:`xml-spec-tags-par` tag. Usually
        set and unset by :py:class:`~imprint.core.tags.ParTag`, but can
        be temporarily switched off and reinstated in response to other
        tags as well. :py:meth:`end_paragraph` deletes this attribute.

    .. py:attribute:: run

       :py:class:`docx.text.run.Run`

        A run is a collection of characters with similar formatting
        within a paragraph. This attribute exists only when parsing a
        :ref:`xml-spec-tags-run` tag. Usually set and unset by
        :py:class:`~imprint.core.tags.RunTag`. :py:meth:`end_paragraph`
        deletes this attribute.

    .. py:attribute:: content

       :py:class:`io.StringIO`

       A mutable buffer used by the engine to accumulate text from the
       :ref:`configuration-xml`.

       Since whitespace needs to be trimmed rather aggressively from
       an XML file, this object gets an extra (non-standard) attribute:

       .. py:attribute:: content.leading_space

          Indicates whether or not to prepend a space when concatenating
          this buffer with others. In general, the text of the first run
          in a paragraph is the only one that does not have this
          attribute set to :py:obj:`True`. This flag is set on the
          buffer rather than the state object itself so that buffers can
          be pushed and popped into the :py:attr:`.content_stack` to
          handle nested tags.

       This attribute should be manipulated mostly through the
       :py:meth:`new_content`, :py:meth:`get_content` and
       :py:meth:`flush_run` methods.

       This attribute must always be present, regardless of the
       position within the document.

       Implemented as a read-write property that can not be deleted or
       set to :py:obj:`None`.

    .. py:attribute:: content_stack

       :py:class:`collections.deque`\ [\ :py:class:`io.StringIO`\ ]

       A stack for nested content buffers. Each buffer represents a tag
       containing independent content. Some tags append to the parent's
       buffer, some close the current buffer to start a new one and
       others, such as :ref:`xml-spec-tags-figure`, use a temporary
       buffer for their content.

       The stack allows for a theoretically indefinite level of nesting
       of text elements. In reality, it will only contain one or two
       elements: the current run text and the contents of interpersed
       tags like :ref:`xml-spec-tags-figure`.

       This attribute should be maniplated through the
       :py:meth:`push_content_stack` and :py:meth:`pop_content_stack`
       methods.

       This attribute may be empty, but never missing. Implemented as a
       read-only property.

    .. py:attribute:: last_list_item

       :py:class:`docx.text.paragraph.Paragraph`

       List items in Word are just paragraphs with a particular style
       and numbering scheme. All of this information can be gathered
       from the previous paragraph that was assigned a concrete list
       numbering instance.

       This attribute should never be missing. It should only be
       :py:obj:`None` to indicate that no prior numbered paragraph has
       occured in the document yet. To this end, it is implemented as a
       read-only property.

    .. py:attribute:: latex_count

       :py:class:`int`

       A counter for the number of :ref:`xml-spec-tags-latex` tags
       encountered so far. Used to generate the file name for the
       equations if :ref:`logging-images` is enabled. Missing otherwise.
    """
    def __init__(self, doc, keywords, references, log):
        """
        Create a new namespace with the specified named arguments.
        """
        references.lock()  # Ensure that this happens no matter what
        super().__init__(references=references, log=log)
        self.__dict__.update(
            doc=doc,
            keywords=keywords,
            content_stack=deque(),
            last_list_item=None,
        )
        self.new_content(False)

    @property
    def doc(self):
        """
        Ensure that :py:attr:`doc` is read-only.
        """
        return self.__dict__['doc']

    @property
    def keywords(self):
        """
        Ensure that :py:attr:`keywords` is read-only.
        """
        return self.__dict__['keywords']

    @property
    def content(self):
        """
        Ensure that :py:attr:`content` is set to a valid
        :py:class:`io.StringIO`.

        Assign a :py:attr:`~content.leading_space` attribute to it if
        not already set.
        """
        return self.__dict__['content']

    @content.setter
    def content(self, value):
        """
        Check the type and ensure that :py:attr:`~content.leading_space`
        is assigned.
        """
        if not isinstance(value, StringIO):
            raise TypeError('Content may only be an io.StringIO object. '
                            'Found {type(value).__name__}')
        if not hasattr(value, 'leading_space'):
            value.leading_space = self.__dict__['content'].leading_space
        self.__dict__['content'] = value

    @property
    def content_stack(self):
        """
        Ensure that :py:attr:`content_stack` is read-only.
        """
        return self.__dict__['content_stack']

    @property
    def last_list_item(self):
        """
        Ensure that :py:attr:`last_list_item` does not get accidentally
        deleted.
        """
        return self.__dict__['last_list_item']

    @last_list_item.setter
    def last_list_item(self, value):
        self.__dict__['last_list_item'] = value

    def new_content(self, leading_space=None):
        """
        Update the :py:attr:`content` text buffer to a new, empty
        :py:class:`~io.StringIO`.

        Calling this method is faster than doing a seek-truncate
        according to http://stackoverflow.com/a/4330829/2988730.

        Parameters
        ----------
        leading_space : tri-state bool
            If :py:obj:`None`, copy :py:attr:`~content.leading_space`
            from the current :py:attr:`content`. Otherwise, set to the
            provided value. The default is to copy the existing value.
        """
        new = StringIO()
        if leading_space is None:
            new.leading_space = self.content.leading_space
        else:
            new.leading_space = bool(leading_space)
        self.content = new

    def get_content(self, default=''):
        """
        Retrieve the text in the current :py:attr:`content` buffer.

        Whitespace is stripped from each line in the text, which is then
        recombined with spaces instead of newlines.

        If the buffer is empty (or contains only whitespace), return
        `default` instead.

        If the text is non-empty, and :py:attr:`content` has
        :py:attr:`~content.leading_space` set to :py:obj:`True`,
        prepended a space.
        """
        text = super().get_content(default)
        # Prepend a leading space to non-empty values that request it.
        if text and self.content.leading_space:
            text = ' ' + text
            #self.content.leading_space = False
                # Pretty sure this should not be done: if content is not
                # removed, neither should the leading white-space be. Possible
                # that it is done for next call to renew not to set the leading
                # whitespace?
        return text

    def flush_run(self, renew=True, default=''):
        """
        Flush the text buffer accumulating the current run into the
        document.

        Text flushing aggressively removes whitespace from around
        individual lines. A single space character is prepended
        before the text if :py:attr:`content.leading_space` is
        :py:obj:`True`.

        If not inside a run, this is a no-op.

        Parameters
        ----------
        renew : bool
            Whether or not to create a new text buffer when finished.
            This is generally a good idea, since the content will
            already be in the document, so the default is
            :py:obj:`True`. The new buffer has
            :py:attr:`~content.leading_space` set to :py:obj:`True`.
        default : str
            The text to insert if the current :py:attr:`content` buffer
            is empty. Defaults to nothing (``''``).
        """
        if 'run' in self:
            text = self.get_content(default)
            if text:
                self.run.add_text(text)
            if renew:
                # Add a space between content, don't add a space otherwise
                space = True if self.run.text else None
                self.new_content(space)

    def push_content_stack(self, flush=False, leading_space=False):
        """
        Temporarily create a new text buffer for the :py:attr:`content`.

        If ``flush`` is :py:obj:`True`, the old buffer is flushed to
        the document and cleared before being pushed to the
        :py:attr:`content_stack`. If ``flush`` is :py:obj:`False`, the
        existing buffer is pushed unchanged. If the content is flushed,
        its :py:attr:`~content.leading_space` attribute is set to
        :py:obj:`True`.

        If the existing buffer is flushed, the buffer that will be
        reinstated when the new one is popped will have
        :py:attr:`~content.leading_space` set to :py:obj:`True`.

        The new buffer can have its :py:attr:`~content.leading_space`
        attribute configured by the ``leading_space`` parameter, which
        defaults to :py:obj:`False`.
        """
        if flush:
            self.flush_run()
        self.content_stack.append(self.content)
        self.new_content(leading_space)
        self.log(logging.TRACE, 'Pushed content stack one level to %d',
                 len(self.content_stack))

    def pop_content_stack(self):
        """
        Reinstate the previous level of the :py:attr:`content_stack` to
        the current :py:attr:`content`.

        Calling this method on an empty stack will cause an error. The
        current :py:attr:`content` is completely discarded.
        """
        self.content = self.content_stack.pop()
        self.log(logging.TRACE, 'Popped content stack one level to %d',
                 len(self.content_stack))

    def check_content_tail(self):
        """
        Include any remaining text in :py:attr:`content` into the last
        run of the last paragraph.

        This ensures that paragraphs get truncated properly, and that
        spurious text between paragraphs is cleaned up.

        A warning is issued if any non-whitepace text is found.
        """
        tail = self.get_content()
        if tail:
            self.log(logging.WARNING, 'Found spurious content outside of run')
            if 'paragraph' in self:
                self.log(logging.TRACE, 'Spurious text within paragraph')
                if self.paragraph.runs:
                    self.log(logging.TRACE, 'Appending to last run')
                    self.paragraph.runs[-1].add_text(tail)
                else:
                    self.log(logging.TRACE, 'Creating new run')
                    self.paragraph.add_run(tail, style=defaults.run_style)
            else:
                self.log(logging.TRACE,
                         'Spurious text outside paragraph: adding a new one')
                paragraph = self.doc.add_paragraph(style=defaults.paragraph_style)
                paragraph.add_run(tail, style=defaults.run_style)
        else:
            self.log(logging.TRACE, 'No content outside run')

        # Make sure the new conent does not have a leading space.
        self.new_content(False)

    def new_run(self, tag, style=defaults.run_style,
                pstyle=defaults.paragraph_style,
                check_in_par=True, keep_par=True):
        """
        Create a new :py:attr:`run`.

        This method handles cases when a run is requested outside a
        paragraph, or inside an existing run:

        - Nested runs are forbidden, but run injection is not.

          - Existing content is flushed for injected runs.

        - Runs outside a paragraph will generate a temporary paragraph
          with a default style.

          - Missing paragraphs can optionally raise a warning.
          - The temporary paragraph can optionally be retained as the
            current paragraph.

        Parameters
        ----------
        name : str
            The name of the tag requesting the run. If there is already
            a :py:attr:`run` attribute present, setting ``name='run'``
            will raise an error because of nesting.
        style : str
            The name of the style to use for the new run.
        pstyle : str
            The name of the style to use for a new paragraph, if one has
            to be created. Moot if there is already a
            :py:attr:`paragraph` attribute.
        check_in_par : bool
            Whether or not to warn if not in a paragraph. Defaults to
            :py:obj:`True`.
        keep_par : bool
            Whether or not to retain a newly created paragraph object in
            the :py:attr:`paragraph` attribute. Moot if there is already
            a :py:attr:`paragraph` attribute.

        Return
        ------
        par : docx.text.paragraph.Paragraph
            The paragraph that the run was added to. If ``keep_par`` is
            :py:obj:`True` or there was already a :py:attr:`paragraph`
            attribute set, this will be the :py:attr:`paragraph`
            attribute.
        run : docx.run.Run
            The newly created run. This will be set to the
            :py:attr:`run` attribute unless there is no existing
            :py:attr:`paragraph` attribute, and ``keep_par`` is set to
            :py:obj:`False`.

        Notes
        -----
        Setting ``keep_par`` to ``False`` for a :ref:`xml-spec-tags-run`
        tag outside a paragraph will cause a situation where
        :py:attr:`run` is set but :py:attr:`paragraph` is not. This may
        cause a problem for the engine, but should never arise with the
        builtin parsers.
        """
        if 'run' in self:
            if tag is not None and tag.casefold() == 'run':
                msg = 'Nested runs are forbidden!'
                self.log(logging.ERROR, msg)
                raise SyntaxError(msg)
            else:
                self.log(logging.DEBUG, 'Run injection detected: <%s>', tag)
                self.flush_run()
        else:
            if 'paragraph' in self:
                space = not is_paragraph_empty(self.paragraph)
            else:
                space = False
            self.new_content(space)

        if 'paragraph' not in self:
            if check_in_par:
                self.log(logging.WARNING, 'Found <%s> outside of <par>', tag)
            paragraph = self.doc.add_paragraph(style=pstyle)
            if keep_par:
                self.paragraph = paragraph
        else:
            paragraph = self.paragraph
        run = paragraph.add_run(style=style)
        if paragraph is self.paragraph:
            self.run = run
        return paragraph, run

    def end_paragraph(self, tag=None):
        """
        Terminate the current paragraph.

        Any existing run is immediately terminated. Spurious text is
        appended to the last available run. Both :py:attr:`paragraph`
        and :py:attr:`run` attributes are deleted by this method.

        If there is no paragraph to terminate, this method is
        equivalent to calling :py:meth:`check_content_tail`.

        Parameters
        ----------
        tag : str or None
            The name of a tag that interrupts the paragraph. If present,
            a warning will be issued. If omitted, no warning will be
            issued.
        """
        self.check_content_tail()
        if 'paragraph' in self:
            if is_paragraph_empty(self.paragraph):
                self.log(logging.DEBUG, 'Found unused paragraph. '
                         'Removing completely.')
                delete_paragraph(self.paragraph)

            del self.paragraph

            if tag is not None:
                self.log(logging.WARNING, '<%s> found in paragraph. '
                         'Paragraph terminated', tag)

        if 'run' in self:
            # Remember that check_content_tail has already appended to the run.
            self.log(logging.WARNING,
                     'Suspicious unterminated run closed with paragraph')
            del self.run

        self.log(logging.TRACE, 'Ended paragraph')

    def insert_picture(self, img, flush_existing=True,
                       style=defaults.figure_run_style,
                       pstyle=defaults.figure_paragraph_style,
                       **kwargs):
        """
        Insert an image into the current document.

        Images must be inserted into a run, so the following cases are
        recognized:

        Outside :ref:`xml-spec-tags-par`
            Create a new temporary
            :py:class:`~docx.text.paragraph.Paragraph` and a new
            :py:class:`~docx.text.run.Run`. Neither object is retained
            (i.e. in :py:attr:`paragraph` and :py:attr:`run`).
        Inside :ref:`xml-spec-tags-par` but outside :ref:`xml-spec-tags-run`
            Create a new temporary :py:class:`~docx.text.run.Run`, which
            will not be retained.
        Inside :ref:`xml-spec-tags-run`
            If the requested ``style`` matches the style of the current
            :py:attr:`run`, it will be flushed and extended. Otherwise,
            the current :py:attr:`run` will be interrupted by a
            temporary run with the new style, and then reinstated.

        It is an error to have a run outside a paragraph.

        Parameters
        ----------
        img : str or file-like
            The image can be the name of a file on disk, or an open file
            (including in memory files like :py:class:`io.BytesIO`). In
            the latter case, the file pointer must be at the beginning
            of the image data.
        style : str
            The name of the
            :doc:`dev/analysis/features/styles/character-style` to apply
            to a new run.
        pstyle : str
            The name of the
            :doc:`dev/analysis/features/styles/paragraph-style` to apply
            if a new paragraph needs to be created.


        Two additional keyword-only arguments can be supplied to
        :py:meth:`~docx.text.run.Run.add_picture`: ``width`` and
        ``height``.
        """
        with self.temp_run(style=style, pstyle=pstyle, keep_same=True):
            pic = self.run.add_picture(img, **kwargs)
        self.log(logging.DEBUG, 'Inserted %s"x%s" picture (WxH)',
                 pic.width.inches, pic.height.inches)

    def inject_par(self, style=defaults.run_style,
                   pstyle=defaults.paragraph_style, text=''):
        """
        Insert a new paragraph into the document with the specified
        styles and text, and return it.

        The contents of the paragraph will be a single run with the
        specified text. Any previously existing :py:attr:`paragraph` and
        :py:attr:`run` will be terminated (see :py:meth:`end_paragraph`)
        and reinstated with their proir styles once the new content is
        inserted.

        Parameters
        ----------
        style : str
            The name of the character style to use for the inserted run.
        pstyle : str
            The name of the paragraph style to apply to the new
            paragraph.
        text : str
            The optional text to insert into the new run.

        Return
        ------
        par : docx.text.paragraph.Paragraph
            The newly created paragraph. This will be a temporary object
            that is never set as :py:attr:`paragraph`.
        run : docx.run.Run
            The newly created run. This will be a temporary object that
            is never set as :py:attr:`run`.
        """
        with self.interrupt_paragraph():
            self.log(logging.DEBUG, 'Injecting par(pstyle=%s, style=%s)',
                     pstyle, style)
            par = self.doc.add_paragraph(style=pstyle)
            run = par.add_run(style=style, text=text)
        return par, run

    @contextmanager
    def temp_run(self, style=defaults.run_style,
                 pstyle=defaults.paragraph_style, keep_same=False):
        """
        Create a temporary run in the current context.

        The run and paragraph styles will be preserved after the context
        manager exits. If the run is injected outside a paragraph, a
        temporary paragraph will be created and forgotten.

        Within the context manager, both :py:attr:`paragraph` and
        :py:attr:`run` are guaranteed to be set to be set. :py:attr:`run`
        will have the style named by ``style``, but :py:attr:`paragraph`
        will only have the style named by ``pstyle`` if it is a
        temporary paragraph.

        All content is flushed into the temporary run when this manager
        exits.

        Parameters
        ----------
        style : str
            The style of the new run.
        pstyle : str
            The style of a new paragraph to contain the run. Used only
            if :py:attr:`paragraph` is unset.
        keep_same : bool
            If :py:obj:`True`, and a run already exists, and has the
            same style as this one, retain it instead of making a new
            one. If :py:obj:`False` (the default), always create a new
            run.
        """
        # Retain old paragraph and run objects
        old_run = self.run if 'run' in self else None
        old_par = self.paragraph if 'paragraph' in self else None

        # Check that keep_same is not activated
        if keep_same and old_run is not None and old_run.style.name == style:
            self.log(logging.DEBUG, 'Not injecting run(style=%s): same style '
                     'as existing', style)
            self.flush_run()
        else:
            if 'run' in self:
                msg = f'into run(style={old_run.style.name})'
            else:
                msg = 'by itself'
            self.log(logging.DEBUG, 'Injecting run(style=%s) %s', style, msg)
            self.new_run(None, style=style, pstyle=pstyle,
                         check_in_par=False, keep_par=True)

        # Do context-manager stuff
        yield

        # Reinstate prior styles
        self.flush_run()
        if self.paragraph is not old_par:
            if 'run' in self:
                del self.run
            self.end_paragraph()
            if old_par is not None:
                self.paragraph = self.doc.add_paragraph(style=old_par.style)
                if old_run is not None:
                    self.run = self.paragraph.add_run(style=old_run.style)
                    self.new_content(False)
        else:
            if old_run is None:
                del self.run
            elif self.run is not old_run:
                self.run = self.paragraph.add_run(style=old_run.style)


    @contextmanager
    def interrupt_paragraph(self, warn=None):
        """
        A context manager for interrupting the current run/paragraph
        and resuming it when complete.

        The current paragraph and run are ended before the body of the
        ``with`` block executes. They are reinstated afterwards, if they
        existed to begin with, with the same styles as before.

        Parameters
        ----------
        warn : str, bool or None
            If a boolean, determines whether or not to issue a generic
            warning if a paragraph is actually interrupted. If a string,
            it is interpreted as the name of the tag that is
            interrupting the paragraph, and mentioned in the warning. No
            warning will be issued if falsy. Defaults to :py:obj:`None`.
        """
        if 'paragraph' in self:
            if 'run' in self:
                style = (self.paragraph.style, self.run.style)
                what = 'run'
                where = '(pstyle=%s, rstyle=%s)'
            else:
                style = (self.paragraph.style,)
                what = 'paragraph'
                where = '(pstyle=%s)'

            if not warn:
                level = logging.DEBUG
                prefix = ''
            else:
                level = logging.WARNING
                prefix = '' if warn is True else f'{warn} inside {what}: '

            self.log(level, '%sInterrupting %s ' + where,
                     prefix, what, *(s.name for s in style))
            self.end_paragraph()
        else:
            style = ()
            self.log(logging.DEBUG, 'Not interrupting anything')

        yield

        if style:
            insert = f'after {warn}' if warn and not isinstance(warn, bool) \
                        else ''
            self.log(logging.TRACE, 'Reinstating paragraph %s with style=%r',
                     insert, style[0])
            self.paragraph = self.doc.add_paragraph(style=style[0])
            if len(style) > 1:
                self.log(logging.TRACE, 'Reinstating run %sstyle=%r', style[1])
                self.run = self.paragraph.add_run(style=style[1])

    def number_paragraph(self, list_type, level):
        """
        Turn the current paragraph into a list item, and store it into
        :py:attr:`last_list_item`.

        The exact numbering scheme depends on :py:attr:`last_list_item`,
        which will be updated to refer to the current paragraph when
        this method completes.

        The following behaviors occur in response to ``list_type``:

        +--------------------------------+---------------------------------+
        | ``list_type``                  | Behavior                        |
        +================================+=================================+
        | :py:obj:`None`                 | Not a list paragraph. Do not    |
        |                                | set numbering or change         |
        |                                | :py:attr:`last_list_item`.      |
        +--------------------------------+---------------------------------+
        | :py:attr:`~ListType.CONTINUED` | Same type and numbering as      |
        |                                | :py:attr:`last_list_item`.      |
        |                                | Set :py:attr:`last_list_item`.  |
        +--------------------------------+---------------------------------+
        | :py:attr:`~ListType.NUMBERED`  | Start a new numbered list.      |
        |                                | Set :py:attr:`last_list_item`.  |
        +--------------------------------+---------------------------------+
        | :py:attr:`~ListType.BULLETED`  | Start a new numbered list.      |
        |                                | Set :py:attr:`last_list_item`.  |
        +--------------------------------+---------------------------------+

        Parameters
        ----------
        list_type : ListType or None
            The type of list to number with, if at all.
        level : int or None
            The depth of the list indentation. :py:obj:`None` means to
            follow the level of the previous list item, if any, or use
            zero depth.
        """
        if list_type is None or 'paragraph' not in self:
            return

        if list_type is ListType.CONTINUED:
            self.log(logging.TRACE, 'Continuing existing list')
            prev = self.last_list_item
            num = None
            if prev is None:
                self.log(logging.WARNING, 'Attempting to continue a list '
                         'when no list exists')
                num = True
        else:
            self.log(logging.TRACE, 'Starting %s list', list_type.value)
            prev = None
            if list_type is ListType.NUMBERED:
                num = True
            elif list_type is ListType.BULLETED:
                num = False
            else:
                raise ValueError(f'Unsupported ListType {list_type!r}')

        list_number(self.doc, self.paragraph, prev, level, num)
        self.last_list_item = self.paragraph

    def image_log_name(self, id, ext=''):
        """
        Create an output name to log an image (or data), for a
        :ref:`plugins-data-configuration` with the given ID, and an
        optional extension.

        This is the standard name-generator for any component (
        :ref:`tag descriptor <tag-api-descriptors>` or
        :ref:`plugin handler <plugins-handlers>`) that enables image
        logging in response to :ref:`keywords-system-log_images`.

        The base name is the result of concatenating an extension-less
        :ref:`keywords-system-log_file` (or
        :ref:`keywords-system-output_docx` if not set), with ``id``,
        separated by an underscore. ``ext`` is appended as-is, if
        provided.
        """
        # select the log file if possible, the output document if not.
        base = self.keywords.get('log_file')
        if not isinstance(base, str):
            base = self.keywords['output_docx']
        path = abspath(base)
        fname, _ = splitext(path)
        return f'{fname}_{id}{ext}'



# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
The :py:mod:`imprint.core` package contains the Imprint
:ref:`introduction-layers-engine`. The :py:mod:`~imprint.core.tags` and
:py:mod:`~imprint.core.state` modules implement most of the
functionality useful to end-users through the public :ref:`tag-api`. The
:py:mod:`~imprint.core.parsers` and :py:mod:`~imprint.core.utilities`
contain the :ref:`internals`.
"""

__all__ = [
    'parsers', 'state', 'tags', 'utilities',
    'root_tag', 'encoding', 'KnownError'
]


#: Name of the root tag expected in XML documents.
root_tag = 'imprint-template'


#: The expected encoding of XML documents
encoding = 'utf-8'


class KnownError(Exception):
    """
    A custom exception class that is used by the engine to indicate that
    a tag or plugin handler exited for a known reason.

    In cases where this exception is logged, the message is printed
    without a stack trace.
    """
    pass



# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
The :py:mod:`imprint.core.tags` module implments the base
:ref:`tag-api`, as well as the all the predefined
:ref:`tag-api-builtins` and :ref:`tag-api-refs`.
"""

from collections.abc import Mapping
import logging
from pprint import pformat
import re

from docx.enum.text import WD_BREAK
from docx.shared import Length

from haggis.files.docx import add_section, insert_toc, str2length
from haggis.latex_util import render_latex

from . import KnownError, utilities
from . import defaults
from .state import ListType


__all__ = [
    'tag_registry', 'TagDescriptor', 'ReferenceDescriptor', 'BuiltinTag',
    'get_key', 'get_handler', 'get_size', 'compute_styles', 'compute_size',
]


class _TagRegistry(dict):
    """
    Implements a limited mapping type to support the
    :py:data:`tag_registry` attribute.

    See the documentation of :py:data:`tag_registry` for a full
    description. Forbidden operations raise a :py:exc:`TypeError`.
    """
    def __delitem__(self, *args, **kwargs):
        """
        Forbidden operation.
        """
        self._delete()

    def __setitem__(self, tag, *args, **kwargs):
        """
        Register a descriptor to a specified tag name.

        Raise a :py:exc:`KeyError` if `tag` is already registered.
        """
        self._overwrite_check(tag)
        super().__setitem__(tag, *args, **kwargs)
        self._add_reference(tag, *args, **kwargs)

    def clear(self, *args, **kwargs):
        """
        Forbidden operation.
        """
        self._delete()

    def pop(self, tag, *args, **kwargs):
        """
        Operation allowed *only* if a `default` is supplied *and* the
        tag is not in the dictionary.
        """
        if tag in self:
            self._delete()
        super().pop(tag, *args, **kwargs)

    def popitem(self, *args, **kwargs):
        """
        Forbidden operation.
        """
        self._delete()

    def setdefault(self, tag, *args, **kwargs):
        """
        Operation always allowed.
        """
        self._string_check(tag)
        # Always safe: never overwrites an existing value
        item = super().setdefault(*args, **kwargs)
        self._add_reference(tag, item)
        return item

    def update(self, other):
        """
        Allowed only if none of the keys in `other` match existing
        keys.
        """
        # Checking for `keys` is how Python duck-types dicts apparently
        # This is fine because dict(other) raises the same errors as
        # update(other).
        if not hasattr(other, 'keys'):
            other = dict(other)
        for tag in other:
            self._overwrite_check(tag)
        super().update(other)
        for item in other.items():
            self._add_reference(*item)

    @property
    def referable_tags(self):
        """
        Compute a list of the names of tags that have a
        non-\ :py:obj:`None` :py:attr:`~TagDescriptor.reference`
        attribute.
        """
        return [key for key, desc in self.items()
                    if TagDescriptor.wrap(desc).reference is not None]

    def _delete(self):
        """
        Utility function for raising an error when any operation
        involving key removal is invoked.
        """
        raise TypeError('Tag deletion not supported')

    def _overwrite_check(self, tag):
        """
        Utility function for raising an error when any operation
        involving key overwrite is invoked.

        This also efficiently performs :py:meth:`_string_check` (i.e.,
        if the tag is not found).
        """
        if tag in self:
            raise KeyError(f'Tag duplication not supported: {tag!r}')
        self._string_check(tag)

    def _string_check(self, tag):
        """
        Utility function for raising an error on non-string keys.
        """
        if not isinstance(tag, str):
            raise TypeError(f'Invalid key type {type(tag).__name__}')

    def _add_reference(self, tag, item):
        """
        Invokes the item's :py:class:`ReferenceDescriptor`\ 's
        registration hook, if the item is referenceable.

        The registration hook may register additional items, so should
        be called after any requested modifications to the registry are
        complete.
        """
        item = TagDescriptor.wrap(item)
        if item.reference is not None:
            item.reference.register(self, tag, item)


#: A limited mapping type that contains all the currently registered tag
#: descriptors.
#:
#: Registering a new descriptor is as easy as doing::
#:
#:     tag_registry[name] = descriptor
#:
#: The registry is a restricted mapping type that supports adding new
#: elements only if they are not already registered. Existing elements
#: can not be deleted. Deletion operations will raise a
#: :py:exc:`TypeError`, while overwriting existing keys will raise a
#: :py:exc:`KeyError`. Aside from that, all operations supported by
#: :py:class:`dict` are allowed (including things like
#: :py:meth:`~dict.update`).
#:
#: Any tag that is :term:`referenceable` by design (has a valid
#: :py:attr:`~TagDescriptor.reference` attribute) will have the
#: :py:class:`ReferenceDescriptor`\ 's registration hook invoked after
#: the tag-proper is registered.
#:
#: The built-in tags are registered when the current module is imported.
#:
#: .. py:attribute:: referable_tags
#:
#:    A convenience property to compute a list of all the keys that have
#:    a non-\ :py:obj:`None` :py:attr:`~TagDescriptor.reference`
#:    attribute in their values.
tag_registry = _TagRegistry()


class TagDescriptor:
    """
    The basis of the tag API.

    Instances of this class contain the information required to process
    a custom tag. They *must* contain all of the attributes listed below,
    with the expected types. The elements in :py:data:`tag_registry` may
    be delegate objects that supply only part of the attibute set. In
    that case, they are wrapped in a proxy as needed at runtime, never
    up-front. The reason for this is twofold:

    1. There may be stateful objects registered for multiple tags, and
       wrapping in a proxy will not allow the tags to share state. This
       would not be a problem, except it would be unexpected behavior.
    2. Some of the attributes may be dynamic properties (or other
       descriptors). Fixing the value once would completely defeat such
       behavior.

    Creating an occasional wrapper around a delegate is not expected to
    be particularly expensive, even if it had to be done for every tag
    encountered in the XML file. On the other hand, it allows for some
    *very* flexible behaviors. At the same time, very few instances of
    wrapping should occur, since most tags will be implemented by
    extending this class and implementing it properly. The
    :py:meth:`wrap` method ensures that all extensions are passed
    through as-is.

    All the :ref:`tag-api-builtins` are instances of children of this
    class.

    .. py:attribute:: content

       A tri-state :py:class:`bool` flag indicating whether the tag is
       allowed/expected to have textual content or not. The values are
       interpreted as follows:

       None
           The tag may not have any content. It must be of the form
           ``<tag/>`` or ``<tag><otherTag>...</otherTag></tag>``.
           Anything else will raise a :term:`fatal error`. If
           :py:attr:`tags` is set to :py:obj:`False`, only the former
           form is allowed.
       False
           The tag *should* not have content, but content will not raise
           an error. A warning will be raised instead.
       True
           The tag is expected to have content, but the content may be
           empty.

       Any value is allowed in a delegate. If defined, the value will be
       converted to :py:class:`bool` if it is not :py:obj:`None`.
       Defaults to :py:obj:`None` if not defined.

    .. py:attribute:: tags

       A :py:class:`bool` indicating whether or not nested tags are
       allowed within this one.

       Any value is allowed in a delegate. If defined, the value will be
       converted to :py:class:`bool`. Defaults to :py:obj:`True` if not
       defined.

    .. py:attribute:: required

       A :py:class:`tuple` of strings containing the name of required
       tag attributes. A tag encountered without all of these attributes
       will raise an error.

       In a delegate, this may be a single string, an iterable of
       strings, :py:obj:`None` or simply omitted. Every element of an
       iterable must be a string, or a :py:exc:`TypeError` is raised
       immediately during construction. Defaults to an empty
       :py:class:`tuple` if not defined.

    .. py:attribute:: optional

       A dictionary mapping the names of optional attributes to their
       default values. Optional attributes are ones that are expected to
       be present in processing, but have sensible defaults that can be
       used, meaning that they do not have to be specified explicitly in
       the :ref:`configuration-xml`.

       In a delegate, this may be any mapping type, an iterable of
       strings, a single string, :py:obj:`None` or simply omitted. In
       the case of an iterable or individual string, all the defaults
       will be :py:obj:`None`. Iterables and mapping keys must be
       strings, or a :py:exc:`TypeError` will be raised 
       during contruction. Defaults to an empty :py:class:`dict` if not
       defined.

    .. py:attribute:: data_config

       The name of the attribute containing the data configuration name
       for the tag. This should only be provided for tags that require
       :ref:`tag-api-configuration-data`. If provided, this tag will
       automatically be added to the :py:attr:`required` sequence.

       In a delegate, this object must be an instance of :py:class:`str`
       or :py:obj:`None`. Defaults to :py:obj:`None` if not defined.

    .. py:attribute:: reference

       A :py:class:`ReferenceDescriptor` that is only present if this
       type of tag can be the target of a reference.

       Examples of referrable built-in tags are
       :ref:`xml-spec-tags-figure`, :ref:`xml-spec-tags-table` and
       sometimes :ref:`xml-spec-tags-par`. Referrable tags can have an
       optional ``role`` attribute that changes the type of reference
       they represent. See the :ref:`tag-api-references-roles`
       description for more information.

       In a delegate, this object must be an instance of
       :py:class:`ReferenceDescriptor` or :py:obj:`None`. Defaults to
       :py:obj:`None` if not defined.
    """
    def __new__(cls, *args, **kwargs):
        """
        Create an empty instance, with all required attributes set to
        default values.

        This method is provided to allow bypassing the default
        :py:meth:`__init__` in child classes. All arguments are ignored.
        """
        self = super().__new__(cls)
        self.content = None
        self.tags = True
        self.required = ()
        self.optional = {}
        self.data_config = None
        self.reference = None
        return self

    def __init__(self, delegate):
        """
        After completion, this instance has all of the required
        attributes defined in the delegate, wrapped in the required
        types.

        A reference to the delegate object is not retained. This method
        can be invoked multiple times. It updates the current descriptor
        with the attributes of the delegate, leaving undefined
        attributes in the delegate untouched.
        """
        def _string_check(iterable, label):
            """
            Verifies that all the elements of an iterable are strings.
    
            Raises :py:exc:`TypeError` with ``label`` in the message if
            the check fails.
            """
            for item in iterable:
                if not isinstance(item, str):
                    raise TypeError(f'{label} tag names must be strings: '
                                    f'found {type(item).__name__}')

        content = getattr(delegate, 'content', self.content)
        self.content = content if content is None else bool(content)

        self.tags = getattr(delegate, 'tags', self.tags)

        required = getattr(delegate, 'required', self.required)
        if required is None:
            required = ()
        elif isinstance(required, str):
            required = (required,)
        else:
            # Implicit iterable check
            required = tuple(required)
        _string_check(required, 'required')
        self.required = required

        optional = getattr(delegate, 'optional', self.optional)
        if optional is None:
            optional = {}
        elif isinstance(optional, str):
            optional = {optional: None}
        elif not isinstance(optional, dict):
            if isinstance(optional, Mapping):
                optional = dict(optional)
            else:
                # Implicit iterable check
                optional = dict.fromkeys(optional)
        _string_check(optional, 'optional')
        self.optional = optional

        start = getattr(delegate, 'start', None)
        if start is not None:
            # Override the non-data method descriptor with a bound method
            self.start = start

        end = getattr(delegate, 'end', None)
        if end is not None:
            # Override the non-data method descriptor with a bound method
            self.end = end

        data_config = getattr(delegate, 'data_config', self.data_config)
        if data_config is not None:
            data_config = str(data_config)
            if data_config not in required:
                self.required += (data_config,)
        self.data_config = data_config

        reference = getattr(delegate, 'reference', self.reference)
        if reference is not None:
            if not isinstance(reference, ReferenceDescriptor):
                raise TypeError('Reference may not be a '
                                f'{type(reference).__name__}')
            for attr in reference.identifiers:
                if attr not in self.required and attr not in self.optional:
                    raise ValueError(
                        'All reference identifiers must be registered '
                        f'attributes. Found {attr!r}.'
                    )
        self.reference = reference

    def start(self, state, name, attr, *args):
        """
        Each descriptor should provide a method with this signature to
        process opening tags.

        If implemented, this method **must** accept the
        :ref:`tag-api-engine-state`, a tag name and a :py:class:`dict`
        of attributes. Normally, the tag name is ignored since a separate
        descriptor is registered for each tag.

        Descriptors that have a non-\ :py:obj:`None`
        :py:attr:`data_config` attribute set will receive an additional
        argument containing the :ref:`tag-api-configuration-data`.

        The default implementation just logs itself.
        """
        state.log(logging.TRACE, 'Opening tag <%s>: No processing required',
                  name)

    def end(self, state, name, attr, *args):
        """
        Each descriptor should provide a method with this signature to
        process closing tags.

        If implemented, this method **must** accept the
        :ref:`tag-api-engine-state`, a tag name and a :py:class:`dict`
        of attributes. Normally, the tag name is ignored since a separate
        descriptor is registered for each tag. The attributes are the
        same as those passed to :py:meth:`start`, barring any
        modifications made *in* :py:meth:`start`.

        Descriptors that have a non-\ :py:obj:`None`
        :py:attr:`data_config` attribute set will receive an additional
        argument containing the :ref:`tag-api-configuration-data`.

        The default implementation just logs itself.
        """
        state.log(logging.TRACE, 'Closing tag <%s>: No processing required',
                  name)

    @classmethod
    def wrap(cls, desc):
        """
        Construct a proxy from the descriptor if it isn't already one.

        This method is provided so that when :py:class:`TagDescriptor`
        objects are implemented properly up front, they do not need to
        be wrapped in an additional layer.

        If the input is a delegate, the return value will always be of
        the type that this method was invoked on. However, the type
        check will always be done agains the base
        :py:class:`TagDescriptor` class.
        """
        return desc if isinstance(desc, __class__) else cls(desc)


class ReferenceDescriptor:
    """
    Defines the process for creating :ref:`tag-api-references` and using
    them through the appropriate tag.

    References are made by processing the :ref:`configuration-xml` and
    mapping out any :term:`referenceable` tags using the
    :py:meth:`start` and :py:meth:`end` methods. In the default
    implementation, the reference text is created by the
    :py:meth:`make_reference` method, invoked from :py:meth:`end`.

    :py:meth:`start` and :py:meth:`end` return a boolean value to allow
    custom tags to be processed selectively. A return value of
    :py:obj:`False` from *either* method means that that the specific
    instance of the tag being processed is not a valid reference target.
    Normally both methods always return :py:obj:`True`, but for the
    builtin :ref:`xml-spec-tags-par` tag, for example, an exception must
    be made.

    References are placed into the document by a special
    :py:class:`TagDescriptor`, which is generally registered along with
    the parent tag that contains a :py:class:`ReferenceDescriptor` using
    the :py:meth:`register` method.

    Current references are purely textual, rather having a dynamic field
    assigned to them. This is still a
    :ref:`work in progress <future-work-anchors>`.

    .. py:attribute:: prefix

       The prefix that normally gets prepended to the reference text.
       Used by :py:meth:`make_reference` to construct the output string.
       Extensions are welcome to ignore this attribute.

    .. py:attribute:: identifiers

       A string or iterable of strings that lists the attributes that
       are used to identify target for this reference type. The
       attribute may be either required or optional for the target
       tag, but it must be recognized either way. This attribute is used
       to check for attributes on tags with a non-default
       :ref:`xml-spec-attributes-role`. Defaults to ``'id'``.
    """
    def __init__(self, prefix, identifiers='id'):
        if isinstance(identifiers, str):
            identifiers = (identifiers,)
        else:
            identifiers = tuple(identifiers)
            for attr in identifiers:
                if not isinstance(attr, str):
                    raise ValueError(
                        'All reference identifier attributes must be strings: '
                        f'found {type(attr).__name__}'
                    )
        self.__dict__.update(prefix=str(prefix), identifiers=identifiers)

    @property
    def prefix(self):
        """
        Ensure that :py:attr:`prefix` is read-only.
        """
        return self.__dict__['prefix']

    @property
    def identifiers(self):
        """
        Ensure that :py:attr:`identifiers` is read-only.
        """
        return self.__dict__['identifiers']

    def start(self, state, name, role, attr):
        """
        Process the opening tag for a referencable tag.

        The default is to log the tag and its role.

        Returns :py:obj:`True` if the tag is a potential reference
        target, :py:obj:`False` if it is definitely not.
        """
        state.log(logging.TRACE, 'Opening <%s> with role %r', name, role)
        return True

    def end(self, state, name, role, attr):
        """
        Process the closing tag for a referencable tag.

        The default is to add the reference to the appropriate map in
        :py:attr:`~imprint.core.state.ReferenceState.references` by ID,
        based on the ``role``, and log the operation. The attribute
        ``id`` is required.

        The actual reference is created by :py:meth:`make_reference`.

        Returns :py:obj:`True` if the tag is definitely a reference
        target, :py:obj:`False` if not.
        """
        computed_role = name if role is None else role
        reference = self.make_reference(state, computed_role, attr)
        self.set_reference(state, computed_role, 'id', attr['id'], reference)
        state.log(logging.DEBUG, 'Closed <%s> with role %r: '
                  'added reference "%s"', name, role, reference)
        return True

    def register(self, registry, name, descriptor):
        """
        A registration hook that is invoked when the parent
        :py:class:`TagDescriptor` is registered.

        The default implementation registers an additional
        :py:class:`TagDescriptor` under the name ``name + '-ref'``,
        which replaces the ``<name-ref/>`` tag with the formatted
        reference. See :py:class:`ReferenceProcessor`.

        Parameters
        ----------
        registry :
            The tag registry that the parent :py:class:`TagDescriptor`
            is being inserted into. See :py:data:`tag_registry` for
            details on the interface.
        name : str
            The name under which the *parent* tag is being registered.
        descriptor :
            The parent object being registered, not necessarily a
            :py:class:`TagDescriptor`. The :py:meth:`TagDescriptor.wrap`
            method can be used to retreive the corresponding
            :py:class:`TagDescriptor` if necessary.
        """
        registry[name + '-ref'] = ReferenceProcessor.get_instance()

    def make_reference(self, state, role, attr):
        """
        Returns a string refering to the specified tag in the specified
        role.

        Keep in mind that the :py:class:`ReferenceDescriptor` is
        selected based on the role, not necessarily the tag name.
        Therefore, the ``role`` argument should always be the "computed"
        role: the name of the tag should be overriden by the value of
        the attribute, if it was specified.
        """
        state.item_counters[role] += 1
        state.log(logging.TRACE, 'Creating new reference for %r, #%d',
                  role, state.item_counters[role])
        return state.format_heading(
            prefix=self.prefix, sep=defaults.reference_level_sep,
            suffix_sep=defaults.reference_sep, suffix=state.item_counters[role]
        )

    def set_reference(self, state, role, attribute, key, reference,
                      duplicates=False):
        """
        Check that the reference identified by ``key`` does not already
        exist and set it.

        Duplicate reference targets cause an error, unless
        ``duplicates`` is :py:obj:`True`, in which case a warning is
        logged and the new value is discarded.
        """
        combined = (role, attribute, key)
        if combined in state.references:
            if duplicates:
                state.log(
                    logging.WARN, 'Duplicate %s %s %r found for "%s" and now '
                    '"%s"', *combined, state.references[combined], reference
                )
            else:
                msg = 'Duplicate %s %s %r'
                state.log(logging.ERROR, msg, *combined)
                raise ValueError(msg % combined)
        else:
            state.log(logging.INFO, 'Adding new reference '
                      '[role=%s, attr=%s, key=%r] = %r', *combined, reference)
            state.references[combined] = reference


class SegmentReferenceDescriptor(ReferenceDescriptor):
    """
    Extension of :py:class:`ReferenceDescriptor` to accumulate heading
    text and allow references through the ``title`` attribute.

    Used by :ref:`xml-spec-tags-par` tags to create heading references.

    .. py:attribute:: heading_style_name_pattern

       A class-level :py:ref:`regular expression <re-objects>` for
       identifying the :ref:`xml-spec-tags-par` tags that represent
       referenceable headings.
    """
    heading_style_name_pattern = re.compile(r'heading (\d+)')

    def start(self, state, name, role, attr):
        """
        Start accumulating content in addition to the default logging.

        If an actual :ref:`xml-spec-tags-par` tag is encountered (as
        opposed to a tag playing that :ref:`xml-spec-attributes-role`),
        and the heading matches ``Heading \\d+``, the current heading
        is incremented in the state.

        If any heading tag, or any tag with ``role="par"`` is
        encountered, a new reference will be created. Non-heading
        paragraphs with no explicit role are non-referenceable. A
        non-heading paragraph can be made referenceable by explicitly
        setting the :ref:`xml-spec-attributes-role`.

        Keep in mind that the title for a segment reference is
        accumulated from **all** the text in the paragraph. Use
        carefully with non-default tags.
        """
        style = attr['style'] or defaults.paragraph_style
        if name == 'par':
            match = self.heading_style_name_pattern.fullmatch(style.lower())
            if match:
                try:
                    level = int(match.group(1))
                except (ValueError, TypeError) as e:
                    state.log(logging.INFO, 'Spurious invalid heading level %r',
                              level)
                    return False
                state.increment_heading(level)
            elif role is None:
                return False

        state.start_content()
        return super().start(state, name, role, attr)

    def end(self, state, name, role, attr):
        """
        Create a dual reference based on the title and optional ID
        in addition to the default logging.
        """
        computed_role = name if role is None else role
        title = state.end_content()
        reference = self.make_reference(state, computed_role, attr, title)
        if attr['id']:
            self.set_reference(state, computed_role, 'id', attr['id'],
                               reference)
            where = 'to ID and TITLE maps'
        else:
            where = 'to TITLE map only'
        self.set_reference(state, computed_role, 'title', title,
                           reference, duplicates=True)
        state.log(logging.DEBUG, 'Closed <%s> with role %r: added '
                  'reference "%s" %s', name, role, reference, where)
        return True

    def make_reference(self, state, role, attr, title):
        """
        Add the section heading to the usual reference text.
        """
        ref = state.format_heading(
            prefix=self.prefix, sep=defaults.reference_level_sep,
            suffix_sep=defaults.reference_sep, suffix=None
        )
        if title:
            ref = f'{ref}: {title}'
        return ref

    def register(self, registry, name, descriptor):
        """
        Register a :py:class:`SegmentRefProcessor` for the
        :ref:`xml-spec-tags-segment-ref` tag.

        This registration hook uses a fixed name, so can only be called
        once.
        """
        registry['segment-ref'] = SegmentRefProcessor.get_instance()


class BuiltinTag(TagDescriptor):
    """
    The base class of all the built-in :py:class:`TagDescriptor`
    implementations.

    Custom tag implementations are welcome to use this class as a base
    instead of a raw :py:class:`TagDescriptor`.
    """
    def __init__(self, delegate=None, **kwargs):
        """
        Updates the required fields with the keywords that are passed
        in.

        If no delegate object (or :py:obj:`None`) is supplied, bypass
        the default constructor (see :py:meth:`TagDescriptor.__new__`).
        `kwargs` will override any defaults and attributes set by a
        delegate.
        """
        if delegate is not None:
            super().__init__(delegate)
        self.__dict__.update(kwargs)


def get_key(key, attr, data, sentinel=None, default=None):
    """
    Resolve the value of ``key`` with respect to ``attr``, but with
    the option to override by the `data` configuration dictionary.

    If the final value is `sentinel`, return `default` instead.
    Return `default` if `key` is missing entirely as well. Both
    `attr` and `data` must be mapping types that support
    a `get` method.
    """
    value = data.get(key, attr.get(key, sentinel))
    return default if value == sentinel else value


def get_size(attr, data, key='size'):
    """
    Convert a string, number or pre-constructed size to a
    :py:class:`docx.shared.Length` object, using :py:func:`get_key`
    for value resolution.

    Common options for ``key`` are ``'width'`` and ``'height'``.

    Valid units suffixes are ``"``, ``in``, ``cm``, ``mm``, ``pt``,
    ``emu``, ``twip``. Default when no units are specified is inches
    (``"``).
    """
    size = get_key(key, attr, data)
    if size is not None and not isinstance(size, Length):
        size = str2length(size)
    return size


def get_handler(tag, id, attr, data, logger, key='handler'):
    """
    Retrieve and load the :ref:`handler <plugins-handlers>` for the
    specified attribute mapping and data configuration.

    If the handler can not be found, a detailed exception is logged and
    a :py:exc:`~imprint.core.KnownError` is raised.
    """
    handler_name = get_key(key, attr, data)
    try:
        handler = utilities.get_handler(handler_name)
    except TypeError as e:
        raise KnownError(f'Unable to find handler "{handler_name}" '
                         f'for {tag} "{id}"') from e
    logger(logging.TRACE, 'Found %r for %r "%s"', handler, tag, id)
    return handler


def get_and_run_handler(tag, id, attr, data, logger, args, kwargs=None,
                        key='handler'):
    """
    Load and run the :ref:`handler <plugins-handlers>` for the
    specified attribute mapping and data configuration.

    If the handler can not be found, a detailed exception is logged, as
    with :py:func:`get_handler`.

    *All* exceptions that occur during execution are converted into
    :py:exc:`~imprint.core.KnownError`.
    """
    handler = get_handler(tag, id, attr, data, logger, key)
    if kwargs is None:
        kwargs = {}
    try:
        return handler(*args, **kwargs)
    except Exception as e:
        raise KnownError(f'Error processing handler for {tag} "{id}": '
                         f'{str(e)}') from e


def compute_styles(attr, data, defaults):
    """
    Compute the required styles based on `attr` and `data`
    configurations.

    Style keys are taken from the keys of `defaults`, while values
    provide the fallback names used if the keys do not appear in either
    `attr` or `data`. Similarly named keys in `data` will override ones
    in ``attr``.
    """
    return {
        style: get_key(style, attr, data, default=default)
            for style, default in defaults.items()
    }


def compute_size(tag, attr, data, logger, default_width=None,
                 width_key='width', height_key='height'):
    """
    Create a dictionary with keys ``width`` and ``height`` and values
    that are instances of :py:class:`docx.shared.Length`.

    Values are resolved according to the rules of :py:func:`get_key`,
    with ``width_key`` and ``height_key`` as the inputs. String values
    may contain units, and will be parsed according to
    :py:func:`get_size`.

    If neither key is present in either configuration (or present but
    set to :py:obj:`None`), set the the width to `default_width`. If
    that is :py:obj:`None` as well, return an empty dictionary.
    """
    width = get_size(attr, data, width_key)
    height = get_size(attr, data, height_key)
    if width is None and height is None:
        if default_width is None:
            message = '%s size not set explicitly.'
            ret = {}
        else:
            message = '%s size not set explicitly. Default width will be used.'
            ret = {'width': default_width}
        args = tag.title(),
    elif width is None:
        message, args = 'Setting custom %s height to %s"', (tag, height.inches)
        ret = {'height': height}
    elif height is None:
        message, args = 'Setting custom %s width to %s"', (tag, width.inches)
        ret = {'width': width}
    else:
        message = 'Fixed-size figure height to W%s" x H%s"'
        args = (width.inches, height.inches)
        ret = {'width': width, 'height': height}

    logger(logging.TRACE, message, *args)
    return ret


class BreakTag(BuiltinTag):
    """
    Implements the :ref:`xml-spec-tags-break` tag.
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('tags', False)
        super().__init__(delegate, **kwargs)

    def end(self, state, name, attr):
        """
        Insert a page break into the document.
        """
        if 'run' in state:
            state.flush_run()
            state.run.add_break(WD_BREAK.PAGE)
        else:
            state.doc.add_page_break()
        state.log(logging.TRACE, 'End of page break')


class ExprTag(BuiltinTag):
    """
    Implements the :ref:`xml-spec-tags-expr` tag.

    .. warning::

       This descriptor uses :py:func:`eval` to execute arbitrary code
       and assign it to a new keyword. Use with extreme caution!
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('content', True)
        kwargs.setdefault('tags', False)
        kwargs.setdefault('required', ('name',))
        kwargs.setdefault('optional', {'imports': ''})
        super().__init__(delegate, **kwargs)

    def start(self, state, name, attr):
        """
        Begin a new expression.

        This just pushes a new
        :py:attr:`~imprint.core.state.EngineState.content_stack` entry
        in the `state`. All content until the closing tag will be
        evaluated as a set of Python statements.
        """
        state.log(logging.TRACE, 'Beginning expression evaluation')
        state.push_content_stack(flush=False)

    def end(self, state, name, attr):
        """
        Evaluate the expression found inside the tag, and add a new
        entry to the ``state``\ 's
        :py:attr:`~imprint.core.state.EngineState.keywords`.

        The :py:attr:`~imprint.core.state.EngineState.content_stack` will
        be popped.

        All errors in importing and evaluation will be propagated up and
        will terminate the parser.
        """
        keyword = attr['name']
        raw = state.content.getvalue()
        imports = {}
        for item in attr['imports'].split():
            name = item.split('.')[0]
            try:
                imports[name] = __import__(item)
            except ImportError:
                state.log(logging.ERROR, 'Error importing "%s" for expression '
                          '"%s": The following evaluation is likely to crash.',
                          item, keyword, exc_info=True)
                raise
        state.log(logging.DEBUG, 'Evaluating expression for "%s" with imports '
                  '"%s": %s', keyword, attr['imports'], raw)
        try:
            value = eval(raw, imports, state.keywords)
        except:
            state.log(logging.ERROR, 'Error evaluating expression for "%s"',
                      keyword)
            raise
        if keyword in state.keywords:
            state.log(logging.WARNING, 'Keyword "%s" already defined. '
                      'Overwriting value "%s" with "%s"', keyword,
                       state.keywords[keyword], value)
        state.log(logging.DEBUG, 'Setting "%s"="%s"', keyword, value)
        state.keywords[keyword] = value
        state.pop_content_stack()
        state.log(logging.TRACE, 'End of expression evaluation')


class FigureTag(BuiltinTag):
    """
    Implements the :ref:`xml-spec-tags-figure` tag.
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('content', False)
        kwargs.setdefault('tags', False)
        kwargs.setdefault('required', ('id', 'handler'))
        kwargs.setdefault('optional', {
            'style': None,
            'pstyle': None,
            'width': None,
            'height': None,
        })
        kwargs.setdefault('reference', ReferenceDescriptor('Figure'))
        kwargs.setdefault('data_config', 'id')
        super().__init__(delegate, **kwargs)

    def start(self, state, name, attr, config):
        """
        Just log the tag.
        """
        state.log(
            logging.DEBUG, 'Starting %s with ID="%s", Handler="%s", '
            'Config=%s', name, attr['id'], attr['handler'],
            pformat(config, indent=4)
        )
        state.push_content_stack(flush=True)

    def end(self, state, name, attr, config):
        """
        Generate and insert a figure based on the selected handler.

        Figures can appear in a run, a paragraph, or on their own.
        """
        id = attr[self.data_config]

        styles = compute_styles(attr, config, {
            'style': defaults.figure_run_style,
            'pstyle': defaults.figure_paragraph_style
        })
        size = compute_size(name, attr, config, logger=state.log,
                            default_width=defaults.figure_width)

        if state.keywords.get('log_images', False):
            # If dumping to file, generate text file name of figure
            figure = state.image_log_name(id)
            figure = get_and_run_handler(
                name, id, attr, config, logger=state.log,
                args=(config, state.keywords, figure)
            )
            if figure is not None:
                state.log(logging.TRACE, 'Saved figure to %s', figure)
        else:
            # If dumping to memory, generate figure in memory
            figure = get_and_run_handler(
                name, id, attr, config, logger=state.log,
                args=(config, state.keywords)
            )
            state.log(logging.TRACE, 'Generated figure in-memory')

        if figure is None:
            state.log(logging.WARNING, 'Unable to create image for '
                      '%s "%s" (no error)', name, id)
            raise KnownError(f'No error in {name} generator {id}')

        state.log(logging.TRACE, 'Successfully generated %s %r', name, id)
        state.insert_picture(figure, **size, **styles)
        state.log(logging.INFO, 'Successfully inserted image for %s %r',
                  name, id)

        state.pop_content_stack()
        state.log(logging.TRACE, 'Figure generation ends')


class KwdTag(BuiltinTag):
    """
    Implements the :ref:`xml-spec-tags-kwd` tag.
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('content', False)
        kwargs.setdefault('tags', False)
        kwargs.setdefault('required', ('name',))
        kwargs.setdefault('optional', {'format': ''})
        super().__init__(delegate, **kwargs)

    def start(self, state, name, attr):
        """
        Find the value of the keyword in the state's
        :py:attr:`~imprint.core.state.EngineState.keywords` and place it
        into the current
        :py:attr:`~imprint.core.state.EngineState.content`.

        If the keyword is not found, a :py:exc:`KeyError` will be
        raised. If the tag has a ``format`` attribute, it is interpreted
        as a :token:`format_spec`, and used to convert the value. If the
        attribute is not present, the value is converted with a simple
        call to :py:class:`str`.
        """
        state.log(logging.TRACE, 'Beginning keyword replacement')
        keyword = attr['name']
        value = state.keywords[keyword]  # Propagate a KeyError here
        state.log(logging.DEBUG, 'Performing keyword replacement for '
                  '"%s"="%s"', keyword, value)
        string = f'{value:{attr["format"]}}'
        state.content.write(string)


class LatexTag(BuiltinTag):
    """
    Implements the :ref:`xml-spec-tags-latex` tag.
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('content', True)
        kwargs.setdefault('tags', False)
        kwargs.setdefault('optional', {
            'style': None,
            'pstyle': None,
            'size': None,
            'dpi': defaults.latex_dpi,
            'format': defaults.latex_format,
        })
        super().__init__(delegate, **kwargs)

    def start(self, state, name, attr):
        """
        Begin a new LaTeX formula.

        Just push a new
        :py:attr:`~imprint.core.state.EngineState.content_stack` entry
        into `state`. All content until the closing tag is evaluated as
        a LaTeX document.
        """
        state.log(logging.TRACE, 'Beginning latex formula')
        state.push_content_stack(flush=True)
        # Get a fresh stack even if the run already made that happen.
        # Avoid conflict with spurious text.

    def end(self, state, name, attr):
        """
        Convert the equation in the text of the current tag into an
        image using :py:func:`haggis.latex_util.render_latex`, and
        insert the image into the parent tag.

        The parent can be a run or a paragraph. If the requested run
        style does not match the current run, the current run will be
        interrupted by a run containing a new picture with the requested
        style, and resumed afterwards. If there is no run to begin with,
        a new run will be created, but not stored in the
        :py:attr:`~imprint.core.state.EngineState.run` attribute of the
        state.

        Formulas are rendered at 96dpi in JPEG format by default.
        """
        text = state.get_content()
        state.log(logging.DEBUG, 'Appending latex formula "%s"', text)

        args = {'format': attr['format'], 'dpi': attr['dpi']}
        size = attr['size']
        if size is not None:
            args['fontsize'] = int(size)

        if state.keywords.get('log_images', False):
            # If dumping to file, generate text file name of figure
            if 'latex_count' in state:
                state.latex_count += 1
            else:
                state.latex_count = 1
            pic = state.image_log_name(f'LaTeX{state.latex_count:04d}',
                                       f'.{attr["format"]}')
            render_latex(text, pic, **args)
            state.log(logging.TRACE, 'Saved latex to %s', pic)
        else:
            # If dumping to memory, generate figure in memory
            pic = render_latex(text, **args)

        styles = compute_styles(attr, {}, {
            'style': defaults.latex_run_style,
            'pstyle': defaults.latex_paragraph_style,
        })

        if 'run' in state:
            state.run.add_text(' ')

        # pop the content stack before inserting the picture to clear the
        # buffer for the run.
        state.pop_content_stack()

        state.log(logging.TRACE, 'Successfully generated %s image', name)
        state.insert_picture(pic, **styles)
        state.log(logging.INFO, 'Successfully inserted %s image', name)

        state.log(logging.TRACE, 'End of latex formula')


class NTag(BuiltinTag):
    """
    Implements the :ref:`xml-spec-tags-n` tag.
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('tags', False)
        super().__init__(delegate, **kwargs)

    def start(self, state, name, attr):
        """
        Add a line break to the current run.

        If not inside a run, append the break to the last run. Make a
        new run only at the start of a paragraph. Ignore with a warning
        outside of a paragraph.
        """
        state.log(logging.TRACE, 'Found newline')
        if 'paragraph' in state:
            if 'run' in state:
                state.log(logging.TRACE, 'Adding to current run')
                state.flush_run()
                state.content.leading_space = False
                run = state.run
            elif state.paragraph.runs:
                state.log(logging.TRACE, 'Adding to previous run')
                run = state.paragraph.runs[-1]
            else:
                state.log(logging.TRACE,
                          'Adding new run to start of paragraph')
                # Do not retain a reference to this run
                run = state.paragraph.add_run(style=defaults.run_style)
            run.add_break(WD_BREAK.LINE)
        else:
            state.log(logging.WARNING, 'Spurious <%s> tag outside of <par>',
                      name)


class ParTag(BuiltinTag):
    """
    Implements the :ref:`xml-spec-tags-par` tag.
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('content', False)
        kwargs.setdefault('optional', {
            'style': None,
            'id': None,
            'list': None,
            'list-level': None,
        })
        kwargs.setdefault('reference', SegmentReferenceDescriptor('Section'))
        super().__init__(delegate, **kwargs)

    def start(self, state, name, attr):
        """
        Terminate any existing paragraph, flush all text and start a
        new paragraph.

        If the new paragraph is a list item, add the necessary metadata
        to it.

        Issue a warning if an existing paragraph is found.
        """
        if 'paragraph' in state:
            state.log(logging.WARNING, 'Nested paragraphs are forbidden. '
                      'Previous paragraph will be terminated.')
            state.end_paragraph(name)

        list_type = self.check_list(state, attr)
        style = self.compute_paragraph_style(state, attr, list_type)
        state.log(logging.TRACE, 'Beginning paragraph with style "%s"', style)
        state.paragraph = state.doc.add_paragraph(style=style)
        if list_type is not None:
            state.log(logging.TRACE, 'Paragraph is a list item of type %r',
                      list_type.value)
        state.number_paragraph(list_type, attr['list-level'])
        

    def end(self, state, name, attr):
        """
        Terminate the current paragraph.

        See :py:meth:`~imprint.core.state.EngineState.end_paragraph` in
        :py:class:`~imprint.core.state.EngineState`.
        """
        state.end_paragraph()

    def check_list(self, state, attr):
        """
        Validate the ``list`` attribute that is found.

        Log an error if the attribute is invalid, but do not terminate
        processing. The attribute is simply ignored if the list is
        neither numbered, bulleted nor continued.

        Return the type normalized to a
        :py:class:`~imprint.core.state.ListType`, or :py:obj:`None` if
        not a list item. If the type is valid, and ``list-level`` is
        set, it is converted to an integer.
        """
        list_type = attr['list']

        if list_type is None:
            state.log(logging.TRACE, 'Not a list paragraph')
            return None

        try:
            list_type = ListType.convert(list_type)
        except ValueError as e:
            state.log(logging.ERROR, 'Unable to parse list type %r',
                      list_type, exc_info=True)
            return None

        level = attr['list-level']
        if level is None:
            label = '<default>'
        else:
            attr['list-level'] = int(level)
            label = level

        state.log(logging.TRACE, '%s list paragraph, level = %s',
                  list_type.value.title(), label)

        return list_type

    def compute_paragraph_style(self, state, attr, list_type):
        """
        Compute the paragraph style based on whether an explicit style
        is set in the attributes, and whether or not the paragraph is a
        list.

            1. If an explicit style is requested, return it.
               Otherwise:
            2. If the paragraph is *not* a list, return the default
               paragraph style. Otherwise:
            3. If the previous paragraph is a list item in the same list
               (i.e., the current ``list-level`` attribute is non-zero),
               return the style of the previous paragraph. Otherwise:
            4. Return the default list item style.

        Parameters
        ----------
        state : ~imprint.core.state.EngineState
            The state is used to check for the previous item's style in
            case #3.
        attr : dict
            The tag attributes, used to check for an explicitly set
            ``style`` as well as for a style reset with
            ``list-level = 0``.
        list_type : ~imprint.core.state.ListType or None
            The type of the list, if a list at all, as returned by
            :py:meth:`check_list`.
        """
        style = attr['style']
        if style is not None:
            return style

        if list_type is None:
            return defaults.paragraph_style

        if list_type is ListType.CONTINUED:
            return state.last_list_item.style.name

        if list_type is ListType.NUMBERED:
            return defaults.numbered_list_paragraph_style

        if list_type is ListType.BULLETED:
            return defaults.bulleted_list_paragraph_style

        # Catchall that "should never happen"
        raise ValueError(f'Unsupported ListType {list_type!r}')


class RunTag(BuiltinTag):
    """
    Implements the :ref:`xml-spec-tags-run` tag.
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('content', True)
        kwargs.setdefault('optional', {'style': defaults.run_style,})
        super().__init__(delegate, **kwargs)

    def start(self, state, name, attr):
        """
        Create a new run, ensuring that there is a paragraph to go with
        it.

        Creating a run outside a paragraph raises a warning and creates
        a paragraph with a default style. See
        :py:meth:`imprint.core.state.EngineState.new_run`.
        """
        style = attr['style']
        state.log(logging.TRACE, 'Beginning run with style "%s"', style)
        state.new_run(name, style)

    def end(self, state, name, attr):
        """
        Place any remaining text into the current run, and remove
        :py:attr:`~imprint.core.state.EngineState.run` attribute of
        ``state``.
        """
        state.flush_run()
        del state.run
        state.log(logging.TRACE, 'Ended run')


class SectionTag(BuiltinTag):
    """
    Implements the :ref:`xml-spec-tags-section` tag.
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('tags', False)
        kwargs.setdefault('optional', {'orientation': 'portrait'})
        super().__init__(delegate, **kwargs)

    def start(self, state, name, attr):
        """
        Begin a new section in the document, optionally altering the
        page orientation.
        """
        state.log(logging.TRACE, 'Start of new document section')
        orientation = attr['orientation']
        with state.interrupt_paragraph(warn=name):
            add_section(self.doc, orientation)
        state.log(logging.INFO, 'Section started with "%s" orientation',
                  orientation)


class SkipTag(BuiltinTag):
    """
    Implements the :ref:`xml-spec-tags-skip` tag.
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('content', True)
        super().__init__(delegate, **kwargs)


class StringTag(BuiltinTag):
    """
    Implements the :ref:`xml-spec-tags-string` tag.
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('content', False)
        kwargs.setdefault('tags', False)
        kwargs.setdefault('required', ('id', 'handler'))
        kwargs.setdefault('data_config', 'id')
        super().__init__(delegate, **kwargs)

    def start(self, state, name, attr, config):
        """
        Just log the tag.
        """
        state.log(logging.DEBUG, 'Starting String with ID="%s", Handler="%s"',
                  attr['id'], attr['handler'])
        state.push_content_stack(flush=True)

    def end(self, state, name, attr, config):
        """
        Generate a string based on the appropriate handler.

        If the ``log_images`` key is set to a truthy value in
        ``state``\ .
        :py:attr:`~imprint.core.state.EngineState.keywords`,
        the content will also be dumped to a file.
        """
        # Ignore contents of tag completely. Reinstate previous content so that
        # a string will be added to it.
        state.pop_content_stack()

        id = attr[self.data_config]
        string = get_and_run_handler(name, id, attr, config, logger=state.log,
                                     args=(config, state.keywords))

        if string is None:
            state.log(logging.WARNING, 'Unable to create text for '
                      '%s "%s" (no error)', name, id)
            raise KnownError(f'No error in {name} generator {id}')

        state.log(logging.DEBUG, 'Successfully generated %s %r: "%s"',
                  name, id, string)

        if state.keywords.get('log_images', False):
            fname = state.image_log_name(id, '.txt')
            with open(fname, 'wt') as file:
                file.write(string)
            state.log(logging.DEBUG, 'Saved %s %r to "%s"', name, id, fname)

        state.content.write(string)
        state.log(logging.INFO, 'Successfully inserted %s %r', name, id)

        state.log(logging.TRACE, 'String generation ends')


class TableTag(BuiltinTag):
    """
    Implements the :ref:`xml-spec-tags-table` tag.
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('content', False)
        kwargs.setdefault('tags', False)
        kwargs.setdefault('required', ('id', 'handler'))
        kwargs.setdefault('optional', {'style': None})
        kwargs.setdefault('reference', ReferenceDescriptor('Table'))
        kwargs.setdefault('data_config', 'id')
        super().__init__(delegate, **kwargs)

    def start(self, state, name, attr, config):
        """
        Just log the tag.
        """
        state.log(logging.DEBUG, 'Starting Table with ID="%s", Handler="%s"',
                  attr['id'], attr['handler'])
        state.push_content_stack(flush=True)

    def end(self, state, name, attr, config):
        """
        Generate and inserts a table based on the selected handler.

        The handler creates the table directly in the document (unlike
        for figures, where only the final product is inserted). Any
        error that occurs mid-processing leaves a stub table in the
        document in addition to the automatically-inserted alt-text.

        Tables appear on their own, outside any paragraph or run, so if
        a table is nested in a run or paragraph, a warning will be
        issued. Any interrupted run or paragraph resumes after the table
        with their prior styles.
        """
        id = attr[self.data_config]
        style = get_key('style', attr, config, default=defaults.table_style)
        image_log = state.image_log_name(id) if \
            state.keywords.get('log_images', False) else None
        with state.interrupt_paragraph(warn=name):
            get_and_run_handler(
                name, id, attr, config, logger=state.log,
                args=(config, state.keywords, state.doc, style),
                kwargs={'image_log_name': image_log},
            )
        state.pop_content_stack()
        state.log(logging.TRACE, 'Table generation ends')


class TocTag(BuiltinTag):
    """
    Implements the :ref:`xml-spec-tags-toc` tag.
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('content', True)
        kwargs.setdefault('tags', False)
        kwargs.setdefault('optional', {
            'min': defaults.toc_min_level,
            'max': defaults.toc_max_level,
            'style': defaults.toc_title_style,
        })
        super().__init__(delegate, **kwargs)

    def start(self, state, name, attr):
        """
        Create a new TOC.

        Log a warning if the tag appears within a paragraph. Truncate
        the paragraph, and resum with the prior style. The same happens
        to the current run, if there is one.
        """
        state.log(logging.TRACE, 'Stubbing out TOC')
        state.log(logging.DEBUG, 'TOC will cover headings from %d to %d',
                  attr['min'], attr['max'])

        # TOC always has a private content stack
        state.push_content_stack(flush=True)

    def end(self, state, name, attr):
        """
        Terminate and insert the TOC.

        Gather any text that has been acquired into the heading, which
        will be a separate pargraph preceding the TOC.

        If the TOC interrupted an existing paragraph, a new paragraph
        will be resumed with the same style as the original. If a run
        style is present as well, a run will be recreated too.
        """
        with state.interrupt_paragraph(name):
            title = state.get_content()
            if title:
                style = attr['style']
                state.log(logging.DEBUG, 'Title of TOC=%s (using style=%r)',
                          title, style)
                state.doc.add_paragraph(style=style).text = title
            insert_toc(state.doc, attr['min'], attr['max'])
            state.pop_content_stack()
        state.log(logging.TRACE, 'End of TOC stub')


class ReferenceProcessor(BuiltinTag):
    """
    Implements the :ref:`xml-spec-tags-figure-ref` and
    :ref:`xml-spec-tags-table-ref` tags.

    This processor is not registered explicitly. It gets added by all of
    the target tags that use it as part of their registration process.
    Registering this processor under a name that does not end in
    ``'-ref'`` will lead to a runtime error in :py:meth:`resolve`.
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('content', None)
        kwargs.setdefault('tags', False)
        kwargs.setdefault('required', ('id',))
        super().__init__(delegate, **kwargs)

    def end(self, state, name, attr):
        """
        Insert a string with the specified reference into the current
        :py:attr:`~imprint.core.state.EngineState.content`.
        """
        state.log(logging.TRACE, 'Start of %s reference resolution', name)
        element = self.resolve(state, name, attr)
        state.content.write(str(element))

    def resolve(self, state, name, attr):
        """
        Overridable operation for fetching and logging the reference
        that is to be inserted.

        The default is to look up the reference by ``'id'`` in the
        :py:class:`imprint.core.state.EngineState`\ 's.
        :py:attr:`~imprint.core.state.EngineState.references`.

        Used by the default implementation of :py:meth:`end`.
        """
        cname = name.lower()
        if not cname.endswith('-ref'):
            raise ValueError(f'Unable to resolve reference {name}: '
                             'does not end in "-ref"')
        tag = cname[:-4]
        id = attr['id']
        element = state.references[tag, 'id', id]
        state.log(logging.INFO, 'Resolved <%s id="%s"> to "%s"',
                  name, id, element)
        return element

    @classmethod
    def get_instance(cls):
        """
        Returns a quasi-singleton instance of the current class.

        This instance is not exposed directly, but it is registered by
        the built-in referencable tags.
        """
        if '_singleton' not in cls.__dict__:
            cls._singleton = cls()
        return cls._singleton


class SegmentRefProcessor(ReferenceProcessor):
    """
    Implements the :ref:`xml-spec-tags-segment-ref` tag.

    This is a special case of :py:class:`ReferenceProcessor` that allows
    access by both ``title`` and ``id``. It's references always resolve
    to a :ref:`xml-spec-tags-par` tag, or a tag playing that
    :ref:`xml-spec-attributes-role`.
    """
    def __init__(self, delegate=None, **kwargs):
        kwargs.setdefault('required', ())
        kwargs.setdefault('optional', {'id': None, 'title': None})
        super().__init__(delegate, **kwargs)

    def resolve(self, state, name, attr):
        """
        Resolve a segment reference be either text or ID.

        Either the ``id`` or ``title`` tag attribute must be present. If
        both are present, they must resolve to the same heading in the
        document or an error is raised.
        """
        id = attr['id']
        title = attr['title']
        if id is not None:
            element = state.references['par', 'id', id]
            if title is not None:
                element2 = state.references['par', 'title', title]
                if element is not element2:
                    state.log(logging.ERROR, 'Mismatch between '
                              'title %r and id %r', title, id)
                    raise ValueError(f'Mismatch between title {title!r} and '
                                     'id {id!r}')
                state.log(logging.INFO, 'Resolved <%s id="%s" title="%s"> '
                          'to "%s"', name, id, title, element)
            else:
                state.log(logging.INFO, 'Resolved <%s id="%s"> to "%s"',
                          name, id, element)
        elif title is not None:
            element = state.references['par', 'title', title]
            state.log(logging.INFO, 'Resolved <%s title="%s"> to "%s"',
                      name, title, element)
        else:
            state.log(logging.ERROR, '<%s> must be identified '
                      'by either title or ID', name)
            raise ValueError('Either "id" or "title" attribute must '
                             f'be present in <{name}>')

        return element


def _register_builtins():
    """
    Called on module import to register all of the
    :ref:`tag-api-builtins`.

    The current implementation registers all subclasses of
    :py:class:`BuiltinTag` whose names end with ``'Tag'``. Tag names are
    extracted from class names by converting CamelCase to camel-case and
    removing the trailing ``'-tag'``. For example,
    :py:class:`TableTag` gets registered for :ref:`xml-spec-tags-table`.
    """
    def name2tag(name):
        """
        Convert CamelCaseNames to all-lowercase-hyphenated-names.
        """
        return name[0].lower() + ''.join(
            c if c.islower() else '-' + c.lower() for c in name[1:]
        )

    from sys import modules
    self = modules[__name__]
    for name in dir(self):
        if name.endswith('Tag'):
            attr = getattr(self, name)
            if isinstance(attr, type) and attr is not BuiltinTag and \
                    issubclass(attr, BuiltinTag):
                name = name2tag(name)[:-4]
                tag_registry[name] = attr()


_register_builtins()



# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
:py:mod:`imprint.code.defaults` defines the defaults used by the engine,
in the absence of a
:ref:`user-defaults file <future-work-user-defaults>`.

This is the last fallback for all the default values used to style
elements without explicit styles set. All the styles defined here exist
in the baseline document created by `python-docx`_ by default.

Importing this module sets up all the attributes based on the user
defaults file.


.. include:: /link-defs.rst
"""

from docx.shared import Inches


__all__ = [
    'paragraph_style', 'run_style',
    'figure_paragraph_style', 'figure_run_style',
    'figure_label_style', 'figure_width',
    'table_paragraph_style',
    'table_run_style', 'table_style',
    'toc_title_style', 'toc_min_level',
    'toc_max_level',
    'latex_paragraph_style', 'latex_run_style',
    'latex_format', 'latex_dpi',
    'numbered_list_paragraph_style', 'bulleted_list_paragraph_style',
    'alt_paragraph_style', 'alt_run_style',
]

#: Default style to apply to unspecified paragraphs.
paragraph_style = 'Normal'

#: Default style to apply to unspecified runs.
run_style = 'Default Paragraph Font'

figure_paragraph_style = 'Quote'
figure_run_style = 'Default Paragraph Font'
figure_label_style = 'Subtitle'
#: The width to make stand-alone figures occupy on the page if one is
#: not specified explicitly in the configuration with a 'width' key.
figure_width = Inches(6.5)

table_style = 'Table Grid'
table_paragraph_style = 'Quote'
table_run_style = 'Default Paragraph Font'

toc_title_style = 'TOC Heading'
toc_min_level = 1
toc_max_level = 3

latex_run_style = 'Default Paragraph Font'
latex_paragraph_style = 'Quote'
latex_format = 'jpg'
latex_dpi = 96

#: The style that will determine which paragraphs are handled specially
#: as list elements.
numbered_list_paragraph_style = 'List Number'
bulleted_list_paragraph_style = 'List Bullet'

alt_paragraph_style = 'Subtitle'
alt_run_style = 'Subtitle Char'

reference_level_sep = '.'
reference_sep = '-'
reference_depth = 3



# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
An implementation of the custom invoice handlers.

A string handler and a table handler are included.
"""

import pandas as pd

from haggis.files import ensure_extension
from haggis.files.docx import style_row, style_column

from imprint.handlers.table import fixed_width_table
from imprint.handlers.utilities import get_file


def _get_df(config, kwds):
    """
    Loads a formatted dataframe based on the configuration.

    Return
    ------
    df : pandas.DataFrame
        The loaded data. The index is a normalized version of the Label
        column.
    damage : float
        The incurred damage, removed from the dataframe.
    security_deposit : float
        The security deposit, removed from the dataframe.
    """
    file = get_file(config, kwds)
    df = pd.read_csv(file, header=0, index_col=None)
    df.set_index(df['Label'].str.title(), inplace=True)
    if 'Damage' in df.index:
        damage = df.loc['Damage', 'Price']
        df.drop(index='Damage', inplace=True)
    else:
        damage = 0.0
    security_deposit = df.loc['Security Deposit', 'Price']
    df.drop(index='Security Deposit', inplace=True)

    return df, damage, security_deposit


def damage_assessment(config, kwds):
    """
    Generate a sentence about property damage.

    The string is based on the numerical value of the damage, taken from
    the line labeled ``Damage`` in the financial data file. The line is
    optional and defaults to zero.

    The line labeled ``Security Deposit`` is compared against if any
    damage occurred. This line is mandatory.
    """
    df, damage, security_deposit = _get_df(config, kwds)

    if damage:
        start = ('The damage done to rental items was assessed at '
                 f'${damage:0.2f}')
        if damage <= security_deposit:
            return start + (', which has already been deducted from your '
                            'security deposit.')

        return start + (
            '. Since your security deposit was insufficient to cover the '
            f'amount, an additional ${damage - security_deposit:0.2f} has '
            'been billed to your account to make up for the difference.'
        )

    return (
        'No damage to rental items was found, so your security deposit has '
        'been applied to the amount due.'
    )


def invoice_table(config, kwds, doc, style, *, image_log_name=None):
    """
    Generate a table containing the actual invoice.

    The table will be 5 columns x (N + 3) rows, including a header row
    and index column. If damage is non-zero, the number of rows is
    (N + 4), with N being the number of items. Image logging is ignored.

    The data comes from a financial file in CSV format. The CSV must
    contain the columns "Label", "Price" and "Quantity". It must contain
    a row labeled "Security Deposit" and optionally one labeled "Damage".

    Configuration Keys
    ------------------
    file : str
        The name of the financial data CSV file (required).
    formatted : bool
        Whether ``file`` is a format string to be interpolated with the
        keywords, or just a raw string.
    """
    df, damage, security_deposit = _get_df(config, kwds)

    rows = df.shape[0] + 3 + bool(damage)

    table = fixed_width_table(doc, style, rows=rows, alignment='^',
                              col_widths=[0.25, 3.25, 1.0, 1.25, 1.25])

    for col, label in enumerate(('ITEM/SERVICE', 'PRICE',
                                 'QUANTITY', 'TOTAL'), start=1):
        table.cell(0, col).text = label

    for row, (index, item) in enumerate(df.iterrows(), start=1):
        table.cell(row, 0).text = f'{row:d}'
        table.cell(row, 1).text = f'{item["Label"]}'
        table.cell(row, 2).text = f'{item["Price"]:0.2f}'
        table.cell(row, 3).text = f'{item["Quantity"]:0.0f}'
        table.cell(row, 4).text = f'{item["Price"] * item["Quantity"]:0.2f}'

    if damage:
        row += 1
        table.cell(row, 0).text = 'X'
        table.cell(row, 1).text = 'Damages'
        table.cell(row, 2).text = f'{damage:0.2f}'
        table.cell(row, 3).text = '-'
        table.cell(row, 4).text = f'{damage:0.2f}'

    row += 1
    table.cell(row, 0).text = 'D'
    table.cell(row, 1).text = 'Security Deposit'
    table.cell(row, 2).text = f'({security_deposit:0.2f})'
    table.cell(row, 3).text = '-'
    table.cell(row, 4).text = f'-{security_deposit:0.2f}'

    row += 1
    total = (df['Price'].dot(df["Quantity"]) + damage - security_deposit)
    table.cell(row, 0).text = 'T'
    table.cell(row, 4).text = f'{total:0.2f}'

    style_column(doc, 'Cell Text',   table, 1)
    style_column(doc, 'Cell Number', table, 2)
    style_column(doc, 'Cell Number', table, 3)
    style_column(doc, 'Cell Number', table, 4)
    style_row(doc, 'Table Heading',  table, 0)
    style_row(doc, 'Strong',         table, -1)

    if image_log_name:
        with open(ensure_extension(image_log_name, '.csv'), 'w') as file:
            df.to_csv(file, index=False, float_format='%0.2f')
            print(f'Security Deposit,{security_deposit:0.2f}', file=file)
            print(f'Damage,{damage:0.2f}', file=file)



# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
:py:mod:`imprint.handlers.string` is the root package for built-in
:ref:`plugins-handlers` for inserting strings into a document.

All the handlers in this module are compatible with the
:ref:`plugin interface <plugins-strings-signature>` used by the
:ref:`xml-spec-tags-string` tag. This package exposes all the handlers
defined in its submodules.
"""

__all__ = []


from .strings import *
from .strings import __all__ as _all
__all__.extend(_all)
del _all



# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
:py:mod:`imprint.handlers.utilities` contains common utilities for
handlers. Users wishing to write their own handlers may want to use
these functions to facilitate a uniform interface. Existing handlers in
this package use these functions as well.
"""

from collections.abc import Mapping


__all__ = ['get_key', 'get_file', 'normalize_descriptor']


def get_key(config, kwds, key, default=None,
            formatted='formatted', missing_ok=True):
    """
    Retreive the value of `key` from the mapping `config`.

    If `key` does not exist in `config`, return `default` instead.

    If `formatted` is a string, it determines the key name that
    determines whether `key` is a format string or not (default is
    yes). Otherwise, it is interpreted as a boolean directly.

    Parameters
    ----------
    config : dict
        The :ref:`plugins-data-configuration` dictionary to search.
    kwds : dict
        The :ref:`keywords` dictionary to use for replacements if
        `formatted` turns out to be truthy.
    key : str
        The name of the key in `config` containing the required value.
    default :
        The value to return if `key` is missing from `config`.
    formatted : str or bool
        Either the name of the key to get the formatted flag from (if a
        string), or the flag itself. In either case, ignored if the
        value is not a string.
    missing_ok : bool
        If truthy, missing values are replaced by `default`. Otherwise
        a :py:exc:`KeyError` is raised.

    Return
    ------
    The value in `config` associated with `key`, optionally formatted
    with `kwds`.
    """
    if missing_ok:
        value = config.get(key, default)
    else:
        value = config[key]
    if isinstance(value, str) and config.get(formatted, True):
        value = value.format(**kwds)
    return value


def get_file(config, kwds, key='file', default=None,
             formatted='formatted', missing_ok=False):
    """
    A wrapper around :py:func:`get_key` that sets the default key to
    ``'file'`` and forbids missing keys.
    """
    return get_key(config, kwds, key, default, formatted, missing_ok)


def normalize_descriptor(descriptor, key, copy=False):
    """
    If `descriptor` is a mapping, return it as-is; otherwise, turn it
    into a value in a mapping keyed by `key`.

    If the descriptor is returned as-is, it can optionally be copied by
    setting `copy` to :py:obj:`True`.
    """
    if isinstance(descriptor, Mapping):
        return descriptor.copy() if copy else descriptor
    return {key: descriptor}



# -*- coding: utf-8 -*-

# imprint: a program for creating documents from data and content templates
#
# Copyright (C) 2019  Joseph R. Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Author: Joseph Fox-Rabinovitz <jfoxrabinovitz at gmail dot com>
# Version: 13 Apr 2019: Initial Coding


"""
:py:mod:`imprint.core.parsers` implements the parsers used to process the
:ref:`configuration-xml`. These parsers make up the heart of the
:ref:`introduction-layers-engine`.

There are currently two parsers: :py:class:`ReferenceProcessor` and
:py:class:`TemplateProcessor`. Both are instances of
:py:class:`haggis.files.xml.SAXLoggable`. The former creates a table of
reference names/titles/locations/numbers that are used by the the latter.
"""

__all__ = ['ReferenceProcessor', 'TemplateProcessor', 'RootTag']


from collections import deque
import logging
import pprint

from haggis.load import module_as_dict
from haggis.files.xml import SAXLoggable
from haggis.files.docx import is_paragraph_empty, delete_paragraph

from . import root_tag, KnownError
from . import defaults
from .tags import TagDescriptor, tag_registry
from .state import ReferenceState, EngineState


class OpenTagError(Exception):
    """
    Used as a goto+label marker when processing opening tags.

    As per https://stackoverflow.com/a/41768438/2988730 and
    https://docs.python.org/3/faq/design.html#why-is-there-no-goto

    This error is raised to indicate a non-fatal error that prevents
    the closing tag from being processed.
    """
    pass


class TagStackNode:
    """
    A structure for maintaining information about open tags for
    :py:class:`TemplateProcessor`.

    All of the attributes except :py:attr:`warned` are immutable, so
    while tempting, a :py:func:`~collections.namedtuple` can not be used.

    All attributes are passed to the constructor in the same order that
    they are listed here. Only the first two are required.

    .. py:attribute:: name

       The name of the tag, not normalized in any way.

    .. py:attribute:: attr

       A plain :py:class:`dict` containing the
       :py:attr:`~imprint.core.tags.TagDescriptor.required` and
       :py:attr:`~imprint.core.tags.TagDescriptor.optional` attributes
       of the tag. This attribute is mutable and gets passed to both the
       :py:meth:`~imprint.core.tags.TagDescriptor.start` and
       :py:meth:`~imprint.core.tags.TagDescriptor.end` methods of the
       tag descriptor. It is not one of the XML library immutable
       mappings.

    .. py:attribute:: descriptor

       The :py:class:`~imprint.core.tags.TagDescriptor` object for this
       tag. This must always be an actual instance of the class, not a
       delegate object to be wrapped. Defaults to :py:obj:`None`.

    .. py:attribute:: config

       The :ref:`tag-api-configuration-data` dictionary, if the
       :py:attr:`descriptor` calls for one, :py:obj:`None` otherwise (the
       default). If the descriptor has a
       :py:attr:`~imprint.core.tags.TagDescriptor.data_config` attribute
       set but this attribute is :py:obj:`None`, then
       :py:attr:`open_error` must be set to :py:obj:`True`.

    .. py:attribute:: open_error

       Lets the closing tag know that a non-fatal error occurred on
       opening, so the closing tag processor should be ignored. Defaults\
       to :py:obj:`False`.

    .. py:attribute:: warned

       Indicates that a text content warning has already been issued for
       a tag that has a
       :py:attr:`~imprint.core.tags.TagDescriptor.content` flag set to
       :py:obj:`False` when nested text is found. Otherwise remains
       :py:obj:`False`. This attribute can not be set by the user on
       initialization.
    """
    __slots__ = (
        'name', 'attr', 'descriptor', 'config', 'open_error', 'warned',
    )

    def __init__(self, name, attr, descriptor=None, config=None,
                 open_error=False):
        # Assume that unknown tags may have text so that it gets pasted
        # in literally and the tag itself is ignored.
        self.descriptor = TagDescriptor.wrap(descriptor)
        if descriptor is None:
            self.descriptor.content = True
        self.name = name
        self.attr = attr  # Must be passed a copy
        self.config = config
        self.warned = False
        self.open_error = open_error


class TagStack:
    """
    A :py:class:`~collections.deque`\ -based stack that does some basic
    structural checking of the XML.

    .. py:attribute:: stack

       The actual stack :py:class:`~collections.deque`, implemented as
       a read-only property.

    .. py:attribute:: current

       The current node. This is just the rightmost node in the stack,
       or :py:obj:`None` if the stack is empty. Also a read-only
       property.
    """
    def __init__(self):
        self.__dict__['stack'] = deque()

    @property
    def stack(self):
        """
        Ensure that :py:attr:`stack` is read-only.
        """
        return self.__dict__['stack']

    @property
    def current(self):
        """
        Return the last element of :py:attr:`stack` or :py:obj:`None` if
        empty.
        """
        stack = self.stack
        return stack[-1] if stack else None

    def push(self, node):
        """
        Push a node onto the stack.

        The node is expected to be a :py:class:`TagStackNode` object,
        but no actual checking is done.
        """
        self.stack.append(node)

    def pop(self, name):
        """
        Pop a node from the stack.

        Popping from an empty stack raises a :py:exc:`TypeError`.
        If the name of the current XML tag supplied as ``name`` and
        the name of the tag in the stack do not match, a
        :py:exc:`ValueError` is raised.

        Return the popped tag.
        """
        tag = self.stack.pop()
        if tag.name != name:
            self.log(logging.ERROR, 'Closing mismatched </%s> tag '
                     'for opening <%s> tag', name, tag.name)
            raise ValueError(f'Malformed XML: Unclosed <{tag.name}> tag')
        return tag


class DocxParserBase(SAXLoggable):
    """
    Base class that contains common functionality of the XML parsers
    that make up the Imprint :ref:`introduction-layers-engine`.

    This class is only intended to avoid code duplication. It serves
    no-standalone purpose whatsoever.

    The XML structure is encoded in the following attributes:

    .. py:attribute:: tag_stack

       A stack with special methods for entering a tag, exiting a tag,
       etc, with some structural validation. The current tag is always
       available via the ``current`` property. Each tag is pushed as an
       object containing the tag name, its (edited) attributes, whether
       or not it expects content and nested tags, and a flag indicating
       whether or not a warning has been raised for unexpected text if
       not. If the tag gets a data configuration, that will be
       referenced as well.
    """
    def __init__(self):
        self.tag_stack = TagStack()

    def check_attributes(self, tag, attr, descriptor=None):
        """
        Validate attributes against the optional and required
        specifications and convert them from a
        :py:class:`xml.sax.xmlreader.Attributes` to a plain
        :py:class:`dict`.

        The ``role`` attribute gets special handling: unless it is
        mentioned explicitly, it is always considered to be
        quasi-optional. It will appear in the output only if it appears
        in the input. ``role`` should never be a required attribute,
        although there is nothing preventing custom tags from setting it
        as such.

        Parameters
        ----------
        tag : str
            The name of the tag.
        attr : xml.sax.xmlreader.Attributes
            Limited mapping containing the actual defined attributes of
            the tag.
        descriptor : 
            Any object passed in will be wrapped in a
            :py:class:`~imprint.core.tags.TagDescriptor`.

        Return
        ------
        asdict : dict
            The attributes of the tag, with all required items, and
            missing optional items filled in with defaults.
        descriptor : ~imprint.core.tags.TagDescriptor
            The input descriptor wrapped in an instance of
            :py:class:`TagDescriptor`. Existing references are passed
            through.
        """
        asdict = {}
        extra_keys = attr.keys()
        descriptor = TagDescriptor.wrap(descriptor)

        for name in descriptor.required:
            if name in attr:
                extra_keys.remove(name)
                asdict[name] = attr[name]
            else:
                msg = 'Missing required "%s" attribute from <%s> tag'
                args = (name, tag)
                msg, args = self.locate(msg, *args)
                raise ValueError(msg % args)

        for name, default in descriptor.optional.items():
            if name in attr:
                extra_keys.remove(name)
                asdict[name] = attr[name]
            else:
                self.log(logging.TRACE, 'Missing optional "%s" attribute '
                         'from <%s> tag', name, tag)
                asdict[name] = default

        # This gets special handling
        if 'role' in attr:
            asdict['role'] = attr['role']
            extra_keys.remove('role')
            if 'role' in tag_registry:
                parent = TagDescriptor.wrap(tag_registry['role'])
                reference = parent.reference
                if reference is not None:
                    for name in reference.identifiers:
                        if name in extra_keys:
                            extra_keys.remove(name)
                            asdict[name] = attr[name]
                        elif name not in asdict:
                            if name in parent.required:
                                msg = ('Missing required "%s" attribute from '
                                       '<%s role="%s"> tag')
                                args = (name, tag, attr['role'])
                                msg, args = self.locate(msg, *args)
                                raise ValueError(msg % args)
                            else:
                                asdict[name] = parent.optional[name]

        if extra_keys:
            self.log(logging.WARNING, '<%s> tag contains unknown '
                     'attributes %s', tag, extra_keys)

        return asdict, descriptor

    def set_document_locator(self, locator, level=logging.DEBUG):
        """
        Set the locator and log it with ``DEBUG`` level.
        """
        super().set_document_locator(locator, level=level)


class RootTag(TagDescriptor):
    """
    Implement the :ref:`xml-spec-root` tag, regardless of its name.

    The root tag is special because any spurious text found within it
    gets stashed in a special paragraph.
    """
    def __init__(self):
        """
        Set up the root tag to accept everything.
        """
        self.__dict__.update(
            tags=True,
            content=False,
            required=(),
            optional={},
            data_config=None,
            reference=None,
        )


tag_registry[root_tag] = RootTag()


class ReferenceProcessor(DocxParserBase):
    """
    The SAX parser that is responsible for pre-computing all the relevant
    references found within the XML template.

    Relevant references are any :term:`referenceable` tags. This
    processor maintains its own reference counter based on the occurence
    of :ref:`xml-spec-tags-figure`, :ref:`xml-spec-tags-table` and other
    tags within :ref:`xml-spec-tags-par` tags with *Heading* styles.
    """
    def __init__(self, heading_depth):
        """
        Initialize the parser and the maps it computes.
        """
        super().__init__()
        self.state = ReferenceState(tag_registry, self.log, heading_depth)

    def startElement(self, name, attr):
        """
        Check for referencable tags, either through the tag itself or
        through the :ref:`xml-spec-attributes-role` attribute.

        If the tag is referenceable, the
        :py:meth:`~imprint.core.tags.ReferenceDescriptor.start` method of
        the :py:class:`~imprint.core.tags.ReferenceDescriptor` is run.

        A new node is added to the stack, with the
        :py:class:`~imprint.core.tags.ReferenceDescriptor` set as its
        :py:attr:`~TagStackNode.config` attribute.
        """
        descriptor = tag_registry.get(name, None)
        attr, descriptor = self.check_attributes(name, attr, descriptor)

        if descriptor.reference is not None or 'role' in attr:
            # Role always overrides actual tag
            if 'role' in attr:
                role = attr['role']
                if role not in tag_registry:
                    raise ValueError(f'Invalid role {role!r}: no such tag')
                reference = TagDescriptor.wrap(tag_registry[role]).reference
                if reference is None:
                    raise ValueError(f'Invalid role {role!r}: '
                                     'not referenceable')
            else:
                role = None
                reference = descriptor.reference
            try:
                open_error = not reference.start(self.state, name, role, attr)
            except:
                self.log(logging.ERROR, 'Fatal error in start of reference '
                         'target <%s>', name)
                raise
        else:
            reference = None
            open_error = True

        tag = TagStackNode(name, attr, descriptor, config=reference,
                            open_error=open_error)
        self.tag_stack.push(tag)

    def endElement(self, name):
        """
        Run the :py:meth:`~imprint.core.tags.ReferenceDescriptor.end`
        method of the :py:class:`~imprint.core.tags.ReferenceDescriptor`
        for the tag, if any.

        Also perform a rudimentary structural check on the XML when
        popping the tag stack.
        """
        tag = self.tag_stack.pop(name)
        if tag.open_error:
            self.log(logging.DEBUG,
                     'Ignoring closing </%s> tag: not a reference', name)
            return

        try:
            tag.config.end(self.state, name, tag.attr.get('role'), tag.attr)
        except:
            self.log(logging.ERROR, 'Fatal error in end of reference '
                     'target </%s>', name, exc_info=True)
            raise

    def characters(self, content):
        """
        Record or discard the raw text in a tag, depending on whether
        the :py:class:`~imprint.core.tags.ReferenceDescriptor` of the
        currently open tag wants it or not.
        """
        if 'content' in self.state:
            self.state.content.write(content)
            verb = 'Acquired'
        else:
            verb = 'Ignoring'
        self.log(logging.TRACE, '%s content "%s"', verb, content)


class TemplateProcessor(DocxParserBase):
    """
    A parser to handle the entire document structure with the assumption
    that a reference mapping has already been made.

    It processes all registered tags, generates all the content,
    replaces all necessary components such as keywords, strings and
    references.

    Much of the processing is handled by the built-in
    :py:class:`~imprint.core.tags.TagDescriptor`\ s and the
    :py:class:`~imprint.core.state.EngineState`. The parser itself
    performs sanity checking of the XML structure based on the
    requirements specified in the descriptors. In addition to checking
    attributes, content and nested tags, it performs a simplistic form
    of XML validation.

    The engine state does not get direct access to the data
    configuration like it does to the `keywords`. The data configuration
    is maintained directly by this class:

    .. py:attribute:: data_config

       A :py:class:`dict` containing all of the data configuration
       objects (dictionaries) loaded from the appropriate module if
       `keywords` contains a ``'data_config'`` key providing the
       module file name, and :py:obj:`None` otherwise. Only
       document setups that actually use data configuration need to
       provide a configuration module.
    """
    # This parser's methods are broken into thematically grouped sections

    #
    # Startup
    #

    def __init__(self, keywords, doc, references):
        """
        Initialize the parser, and set up an
        :class:`~imprint.core.state.EngineState` for communication.
        """
        super().__init__()

        if 'data_config' in keywords:
            self.data_config = module_as_dict(keywords['data_config'])
        else:
            self.data_config = None

        self.logger.debug("Initialized for %s", doc)

        self.state = EngineState(doc=doc,
                                 keywords=keywords,
                                 references=references,
                                 log=self.log)
        self.state.new_content(False)

        # Trim the document
        if len(doc.tables) == 0 and len(doc.paragraphs) == 1 and \
                is_paragraph_empty(doc.paragraphs[0]):
            delete_paragraph(doc.paragraphs[0])


    #
    # Internal special-case handlers
    #

    def _handle_os_error(self, exc, tag, id=None):
        """
        Handle :py:exc:`FileNotFoundError` and :py:exc:`PermissionError`
        that occur during tag processing in handlers in a non-fatal way.

        These errors are semi-expected so they get logged, but do not
        cause a fatal crash of the program.
        """
        if isinstance(exc, FileNotFoundError):
            op = 'find'
        elif isinstance(exc, PermissionError):
            op = 'access'
        else:
            raise

        if id is None:
            self.log(logging.ERROR, 'Unable to %s dataset "%s" '
                     'in for %s tag', op, exc.filename, tag)
        else:
            self.log(logging.ERROR, 'Unable to %s dataset "%s" '
                     'in handler for %s "%s"', op, exc.filename, tag, id)
        self.log(logging.TRACE, 'Stack Trace:', exc_info=True)

    def _handle_known_error(self, exc, tag, id=None):
        """
        Handle :py:exc:`~imprint.core.KnownError` that occurs during tag
        processing in handlers in a non-fatal way.

        These errors are semi-expected so they get logged, but do not
        cause a fatal crash of the program.
        """
        if id is None:
            self.log(logging.ERROR, 'Non-fatal error in tag %s: %s',
                     tag, str(exc))
        else:
            self.log(logging.ERROR, 'Non-fatal error in tag %s "%s": %s',
                     tag, id, str(exc))
        self.log(logging.TRACE, 'Caused by:', exc_info=True)

    def _insert_alt_text(self, tag, id, reason):
        """
        Insert a run with alt-text into the current state.

        The logged error will contain the reason for the failure to
        generate content.
        """
        # Missing required configuration
        self.log(logging.ERROR, 'Replacing opening %s '
                 '"%s" with alt-text: %s', tag, id, reason)
        with self.state.temp_run(style=defaults.alt_run_style,
                                 pstyle=defaults.alt_paragraph_style):
            self.state.run.text = f'{{ Insert {tag} [{id}] here }}'

    #
    # Actual processing methods
    #

    def startElement(self, name, attr):
        """
        SAX method to process all open tags.
        """
        # Check if nested tags are allowed at all first
        tag = self.tag_stack.current
        if tag and not tag.descriptor.tags:
            msg = 'Nesting tags in <%s> is forbidden: found <%s>'
            args = tag.name, name
            msg, args = self.locate(msg, *args)
            self.log(logging.Error, msg, *args)
            raise ValueError(msg % args)

        self.log(logging.TRACE, 'Starting element <%s> with attributes %s',
                 name, attr.items())

        descriptor = config = config_name = None
        try:
            if name in tag_registry:
                descriptor = tag_registry[name]
            else:
                # Handle unregistered tags
                self.log(logging.WARNING, 'Unrecognized tag <%s> '
                         'will be ignored', name)
                raise OpenTagError
            attr, descriptor = self.check_attributes(name, attr, descriptor)
            start_args = [self.state, name, attr]
            if descriptor.data_config is not None:
                if self.data_config is None:
                    raise ValueError(
                        'No data configuration defined for parser. Please set '
                        'the `data_configutation` key in your master '
                        'configuration file to point to the appropriate data '
                        'configuration file for this document.'
                    )
                # Handle dynamic configurations
                config_name = attr[descriptor.data_config]
                if config_name in self.data_config:
                    config = self.data_config[config_name]
                    start_args.append(config)
                    self.log(
                        logging.DEBUG, 'Configuration for %s %r:\n%s',
                        name, config_name, pprint.pformat(config, indent=4)
                    )
                else:
                    self._insert_alt_text(name, config_name, 'missing '
                                          'data configuration dictionary')
                    raise OpenTagError

            try:
                descriptor.start(*start_args)
            except KnownError as e:
                self._handle_known_error(e, name, config_name)
                raise OpenTagError
            except OSError as e:
                self._handle_os_error(e, name, config_name)
                raise OpenTagError
            except:
                self.log(logging.ERROR, "Fatal error in start tag for <%s>",
                         name)
                if config_name is not None:
                    self._insert_alt_text(name, config_name,
                                          'Error in opening tag processor')
                raise
        except OpenTagError:
            open_error = True
        else:
            open_error = False

        tag = TagStackNode(name, attr, descriptor, config, open_error)
        self.tag_stack.push(tag)

    def endElement(self, name):
        """
        SAX method to process all closing tags.
        """
        self.log(logging.TRACE, 'Ending element %s', name)
        tag = self.tag_stack.pop(name)
        if tag.open_error:
            self.log(logging.INFO, 'Error in opening tag, '
                     'ignoring closing tag </%s>', name)
            return

        end_args = [self.state, name, tag.attr]
        if tag.config is None:
            config_name = None
        else:
            end_args.append(tag.config)
            config_name = tag.attr[tag.descriptor.data_config]

        try:
            tag.descriptor.end(*end_args)
            err = False
        except KnownError as e:
            self._handle_known_error(e, name, config_name)
            err = True
        except OSError as e:
            self._handle_os_error(e, name, config_name)
            err = True
        except:
            self.log(logging.ERROR, 'Fatal error in end tag for </%s>',
                     name, exc_info=True)
            if config_name is not None:
                self._insert_alt_text(name, config_name,
                                      'Error in closing tag processor')
            err = True
            raise
        finally:
            if err and config_name is not None:
                self._insert_alt_text(name, config_name,
                                      'Error in closing tag processor')

    def characters(self, content):
        """
        SAX method to process all raw text.
        """
        self.log(logging.TRACE, 'Acquired some content: "%s"', content)
        stripped = content.strip()
        tag = self.tag_stack.current
        content_allowed = tag.descriptor.content
        if content_allowed is None and stripped:
            self.log(logging.ERROR, '<%s> tag must be empty: found "%s"',
                     tag.name, stripped)
            raise ValueError('invalid content in <%s> tag' % tag.name)
        if not content_allowed and stripped:
            if not tag.warned:
                self.log(logging.WARNING, 'Unexpected content in <%s>',
                         tag.name)
                tag.warned = True
            if tag.name == root_tag:
                self.log(logging.WARNING, 'Since content is in root, '
                         'a new paragraph will be started')
                self.state.new_run(tag.name, check_in_par=False)
        self.state.content.write(content)

    def endDocument(self):
        """
        SAX method to close the document.
        """
        self.log(logging.TRACE, 'Ending XML document parsing')
        self.state.check_content_tail()
        if 'paragraph' in self.state:
            del self.state.paragraph

    #
    # Things we don't care about (but allow to pass)
    #

    def startDocument(self):
        """
        SAX method to start the document.

        Ignored with a log message.
        """
        self.log(logging.TRACE, "Starting XML document parsing")

    def ignorableWhitespace(self, whitespace):
        """
        SAX method to process ignorable whitespace.

        Ignored with a log message.
        """
        self.log(logging.TRACE, "Encountered ignorable whitespace: \"%s\"", ".".join(whitespace))

    #
    # Things we should never see
    #

    def startPrefixMapping(self, prefix, uri):
        """
        SAX method to begin a prefix mapping.

        This should never happen in an Imprint template.
        """
        self.log(logging.WARNING, "Starting prefix mapping for %s:%s", prefix, uri)

    def endPrefixMapping(self, prefix, uri):
        """
        SAX method to end a prefix mapping.

        This should never happen in an Imprint template.
        """
        self.log(logging.WARNING, "Ending prefix mapping for %s:%s", prefix, uri)

    def startElementNS(self, name, qname, attr):
        """
        SAX method to handle an opening tag with a namespace.

        This should never happen in an Imprint template.
        """
        self.log(logging.WARNING, "Starting namespace element %s (%s) with attributes %s",
                   name, qname, attr.items())

    def endElementNS(self, name, qname):
        """
        SAX method to handle a closing tag with a namespace.

        This should never happen in an Imprint template.
        """
        self.log(logging.WARNING, "Ending namespace element %s (%s)", name, qname)

    def processingInstruction(self, target, data):
        """
        SAX method to handle an XML processing instruction.

        This should never happen in an Imprint template.
        """
        self.log(logging.WARNING, "Encountered processing instruction <%s %s>", target, data)

    def skippedEntity(self, name):
        """
        SAX method to process a skipped XML entity.

        This should never happen in an Imprint template.
        """
        self.log(logging.WARNING, "Encountered skipped entity %s", name)
